import re
import binascii
import asyncio
import logging
import typing
import socket
from typing import List, Set, Dict, Tuple, Optional
from asyncio.transports import DatagramTransport
from aioupnp.fault import UPnPError
from aioupnp.serialization.ssdp import SSDPDatagram
from aioupnp.constants import SSDP_IP_ADDRESS, SSDP_PORT
from aioupnp.protocols.multicast import MulticastProtocol
from aioupnp.protocols.m_search_patterns import packet_generator

ADDRESS_REGEX = re.compile("^http:\/\/(\d+\.\d+\.\d+\.\d+)\:(\d*)(\/[\w|\/|\:|\-|\.]*)$")

log = logging.getLogger(__name__)


class PendingSearch(typing.NamedTuple):
    address: str
    st: str
    fut: 'asyncio.Future[SSDPDatagram]'


class SSDPProtocol(MulticastProtocol):
    def __init__(self, multicast_address: str, lan_address: str,
                 loop: Optional[asyncio.AbstractEventLoop] = None) -> None:
        super().__init__(multicast_address, lan_address)
        self.loop: asyncio.AbstractEventLoop = loop or asyncio.get_event_loop()
        self.transport: Optional[DatagramTransport] = None
        self._pending_searches: List[PendingSearch] = []
        self.notifications: List[SSDPDatagram] = []
        self.connected = asyncio.Event(loop=self.loop)
        self.devices: 'asyncio.Queue[SSDPDatagram]' = asyncio.Queue(loop=self.loop)

    def connection_made(self, transport: asyncio.DatagramTransport) -> None:  # type: ignore
        super().connection_made(transport)
        self.connected.set()
        return None

    def disconnect(self) -> None:
        if self.transport:
            self.leave_group(self.multicast_address, self.bind_address)
            self.transport.close()
        self.connected.clear()
        while self._pending_searches:
            pending = self._pending_searches.pop()[2]
            if not pending.cancelled() and not pending.done():
                pending.cancel()
        return None

    def _callback_m_search_ok(self, address: str, packet: SSDPDatagram) -> None:
        futures: Set['asyncio.Future[SSDPDatagram]'] = set()
        replied: List[PendingSearch] = []

        for pending in self._pending_searches:
            # if pending.address == address and pending.st in (packet.st, "upnp:rootdevice"):
            if pending.address == address and pending.st == packet.st:
                replied.append(pending)
                if pending.fut not in futures:
                    futures.add(pending.fut)
        if replied:
            self.devices.put_nowait(packet)

        while replied:
            self._pending_searches.remove(replied.pop())

        while futures:
            fut = futures.pop()
            if not fut.done():
                fut.set_result(packet)

    def _send_m_search(self, address: str, packet: SSDPDatagram, fut: 'asyncio.Future[SSDPDatagram]') -> None:
        if not self.transport:
            if not fut.done():
                fut.set_exception(UPnPError("SSDP transport not connected"))
            return
        assert packet.st is not None
        self._pending_searches.append(
            PendingSearch(address, packet.st, fut)
        )
        self.transport.sendto(packet.encode().encode(), (SSDP_IP_ADDRESS, SSDP_PORT))

        # also send unicast
        log.debug("send m search to %s: %s", address, packet.st)
        self.transport.sendto(packet.encode().encode(), (address, SSDP_PORT))

    def send_m_searches(self, address: str,
                        datagrams: List[Dict[str, typing.Union[str, int]]]) -> 'asyncio.Future[SSDPDatagram]':
        fut: 'asyncio.Future[SSDPDatagram]' = self.loop.create_future()
        for datagram in datagrams:
            packet = SSDPDatagram("M-SEARCH", datagram)
            assert packet.st is not None
            self._send_m_search(address, packet, fut)
        return fut

    async def m_search(self, address: str, timeout: float,
                       datagrams: List[Dict[str, typing.Union[str, int]]]) -> SSDPDatagram:
        fut = self.send_m_searches(address, datagrams)
        return await asyncio.wait_for(fut, timeout, loop=self.loop)

    def datagram_received(self, data: bytes, addr: Tuple[str, int]) -> None:  # type: ignore
        if addr[0] == self.bind_address:
            return None
        try:
            packet = SSDPDatagram.decode(data)
            log.debug("decoded packet from %s:%i: %s", addr[0], addr[1], packet)
        except UPnPError as err:
            log.warning("failed to decode SSDP packet from %s:%i (%s): %s", addr[0], addr[1], err,
                        binascii.hexlify(data))
            return None
        if packet._packet_type == packet._OK:
            self._callback_m_search_ok(addr[0], packet)
            return None
        # elif packet._packet_type == packet._NOTIFY:
        #     log.debug("%s:%i sent us a notification: %s", packet)
        #     if packet.nt == SSDP_ROOT_DEVICE:
        #         address, port, path = ADDRESS_REGEX.findall(packet.location)[0]
        #         key = None
        #         for (addr, service) in self.discover_callbacks:
        #             if addr == address:
        #                 key = (addr, service)
        #                 break
        #         if key:
        #             log.debug("got a notification with the requested m-search info")
        #             notify_fut: Future = self.discover_callbacks.pop(key)
        #             notify_fut.set_result(SSDPDatagram(
        #                 SSDPDatagram._OK, cache_control='', location=packet.location, server=packet.server,
        #                 st=UPNP_ORG_IGD, usn=packet.usn
        #             ))
        #         self.notifications.append(packet.as_dict())
        #         return


async def listen_ssdp(lan_address: str, gateway_address: str,
                      loop: Optional[asyncio.AbstractEventLoop] = None) -> Tuple[SSDPProtocol, str, str]:
    loop = loop or asyncio.get_event_loop()
    try:
        sock: socket.socket = SSDPProtocol.create_multicast_socket(lan_address)
        listen_result: Tuple[asyncio.BaseTransport, asyncio.BaseProtocol] = await loop.create_datagram_endpoint(
            lambda: SSDPProtocol(SSDP_IP_ADDRESS, lan_address), sock=sock
        )
        protocol = listen_result[1]
        assert isinstance(protocol, SSDPProtocol)
    except Exception as err:
        raise UPnPError(err)
    else:
        protocol.join_group(protocol.multicast_address, protocol.bind_address)
        protocol.set_ttl(1)
    return protocol, gateway_address, lan_address


async def m_search(lan_address: str, gateway_address: str, datagram_args: Dict[str, typing.Union[int, str]],
                   timeout: int = 1,
                   loop: Optional[asyncio.AbstractEventLoop] = None) -> SSDPDatagram:
    protocol, gateway_address, lan_address = await listen_ssdp(
        lan_address, gateway_address, loop
    )
    try:
        return await protocol.m_search(address=gateway_address, timeout=timeout, datagrams=[datagram_args])
    except asyncio.TimeoutError:
        raise UPnPError("M-SEARCH for {}:{} timed out".format(gateway_address, SSDP_PORT))
    finally:
        protocol.disconnect()


async def multi_m_search(lan_address: str, gateway_address: str, timeout: int = 3,
                         loop: Optional[asyncio.AbstractEventLoop] = None) -> SSDPProtocol:
    loop = loop or asyncio.get_event_loop()
    protocol, gateway_address, lan_address = await listen_ssdp(
        lan_address, gateway_address, loop
    )
    fut = asyncio.ensure_future(protocol.send_m_searches(
        address=gateway_address, datagrams=list(packet_generator())
    ), loop=loop)
    loop.call_later(timeout, lambda: None if not fut or fut.done() else fut.cancel())
    return protocol



import unittest
from aioupnp.fault import UPnPError
from aioupnp.serialization.scpd import serialize_scpd_get, deserialize_scpd_get_response
from aioupnp.serialization.xml import xml_to_dict
from aioupnp.device import Device
from aioupnp.util import get_dict_val_case_insensitive


class TestSCPDSerialization(unittest.TestCase):
    path, lan_address = '/IGDdevicedesc_brlan0.xml', '10.1.10.1'
    get_request = b'GET /IGDdevicedesc_brlan0.xml HTTP/1.1\r\n' \
                  b'Accept-Encoding: gzip\r\nHost: 10.1.10.1\r\nConnection: Close\r\n\r\n'

    response = b"HTTP/1.1 200 OK\r\n" \
               b"CONTENT-LENGTH: 2972\r\n" \
               b"CONTENT-TYPE: text/xml\r\n" \
               b"DATE: Thu, 18 Oct 2018 01:20:23 GMT\r\n" \
               b"LAST-MODIFIED: Fri, 28 Sep 2018 18:35:48 GMT\r\n" \
               b"SERVER: Linux/3.14.28-Prod_17.2, UPnP/1.0, Portable SDK for UPnP devices/1.6.22\r\n" \
               b"X-User-Agent: redsonic\r\n" \
               b"CONNECTION: close\r\n" \
               b"\r\n" \
               b"<?xml version=\"1.0\"?>\n<root xmlns=\"urn:schemas-upnp-org:device-1-0\">\n<specVersion>\n<major>1</major>\n<minor>0</minor>\n</specVersion>\n<device>\n<deviceType>urn:schemas-upnp-org:device:InternetGatewayDevice:1</deviceType>\n<friendlyName>CGA4131COM</friendlyName>\n<manufacturer>Cisco</manufacturer>\n<manufacturerURL>http://www.cisco.com/</manufacturerURL>\n<modelDescription>CGA4131COM</modelDescription>\n<modelName>CGA4131COM</modelName>\n<modelNumber>CGA4131COM</modelNumber>\n<modelURL>http://www.cisco.com</modelURL>\n<serialNumber></serialNumber>\n<UDN>uuid:11111111-2222-3333-4444-555555555556</UDN>\n<UPC>CGA4131COM</UPC>\n<serviceList>\n<service>\n<serviceType>urn:schemas-upnp-org:service:Layer3Forwarding:1</serviceType>\n<serviceId>urn:upnp-org:serviceId:L3Forwarding1</serviceId>\n<SCPDURL>/Layer3ForwardingSCPD.xml</SCPDURL>\n<controlURL>/upnp/control/Layer3Forwarding</controlURL>\n<eventSubURL>/upnp/event/Layer3Forwarding</eventSubURL>\n</service>\n</serviceList>\n<deviceList>\n<device>\n<deviceType>urn:schemas-upnp-org:device:WANDevice:1</deviceType>\n<friendlyName>WANDevice:1</friendlyName>\n<manufacturer>Cisco</manufacturer>\n<manufacturerURL>http://www.cisco.com/</manufacturerURL>\n<modelDescription>CGA4131COM</modelDescription>\n<modelName>CGA4131COM</modelName>\n<modelNumber>CGA4131COM</modelNumber>\n<modelURL>http://www.cisco.com</modelURL>\n<serialNumber></serialNumber>\n<UDN>uuid:11111111-2222-3333-4444-555555555556</UDN>\n<UPC>CGA4131COM</UPC>\n<serviceList>\n<service>\n<serviceType>urn:schemas-upnp-org:service:WANCommonInterfaceConfig:1</serviceType>\n<serviceId>urn:upnp-org:serviceId:WANCommonIFC1</serviceId>\n<SCPDURL>/WANCommonInterfaceConfigSCPD.xml</SCPDURL>\n<controlURL>/upnp/control/WANCommonInterfaceConfig0</controlURL>\n<eventSubURL>/upnp/event/WANCommonInterfaceConfig0</eventSubURL>\n</service>\n</serviceList>\n<deviceList>\n    <device>\n        <deviceType>urn:schemas-upnp-org:device:WANConnectionDevice:1</deviceType>\n        <friendlyName>WANConnectionDevice:1</friendlyName>\n        <manufacturer>Cisco</manufacturer>\n        <manufacturerURL>http://www.cisco.com/</manufacturerURL>\n        <modelDescription>CGA4131COM</modelDescription>\n        <modelName>CGA4131COM</modelName>\n        <modelNumber>CGA4131COM</modelNumber>\n        <modelURL>http://www.cisco.com</modelURL>\n        <serialNumber></serialNumber>\n        <UDN>uuid:11111111-2222-3333-4444-555555555555</UDN>\n        <UPC>CGA4131COM</UPC>\n        <serviceList>\n       <service>\n           <serviceType>urn:schemas-upnp-org:service:WANIPConnection:1</serviceType>\n           <serviceId>urn:upnp-org:serviceId:WANIPConn1</serviceId>\n           <SCPDURL>/WANIPConnectionServiceSCPD.xml</SCPDURL>\n           <controlURL>/upnp/control/WANIPConnection0</controlURL>\n           <eventSubURL>/upnp/event/WANIPConnection0</eventSubURL>\n       </service>\n        </serviceList>\n    </device>\n</deviceList>\n</device>\n</deviceList>\n<presentationURL>http://10.1.10.1/</presentationURL></device>\n</root>\n"

    response_bad_root_device_name = b"HTTP/1.1 200 OK\r\n" \
               b"CONTENT-LENGTH: 2972\r\n" \
               b"CONTENT-TYPE: text/xml\r\n" \
               b"DATE: Thu, 18 Oct 2018 01:20:23 GMT\r\n" \
               b"LAST-MODIFIED: Fri, 28 Sep 2018 18:35:48 GMT\r\n" \
               b"SERVER: Linux/3.14.28-Prod_17.2, UPnP/1.0, Portable SDK for UPnP devices/1.6.22\r\n" \
               b"X-User-Agent: redsonic\r\n" \
               b"CONNECTION: close\r\n" \
               b"\r\n" \
               b"<?xml version=\"1.0\"?>\n<root xmlns=\"urn:schemas-upnp-org:device-1-?\">\n<specVersion>\n<major>1</major>\n<minor>0</minor>\n</specVersion>\n<device>\n<deviceType>urn:schemas-upnp-org:device:InternetGatewayDevic3:1</deviceType>\n<friendlyName>CGA4131COM</friendlyName>\n<manufacturer>Cisco</manufacturer>\n<manufacturerURL>http://www.cisco.com/</manufacturerURL>\n<modelDescription>CGA4131COM</modelDescription>\n<modelName>CGA4131COM</modelName>\n<modelNumber>CGA4131COM</modelNumber>\n<modelURL>http://www.cisco.com</modelURL>\n<serialNumber></serialNumber>\n<UDN>uuid:11111111-2222-3333-4444-555555555556</UDN>\n<UPC>CGA4131COM</UPC>\n<serviceList>\n<service>\n<serviceType>urn:schemas-upnp-org:service:Layer3Forwarding:1</serviceType>\n<serviceId>urn:upnp-org:serviceId:L3Forwarding1</serviceId>\n<SCPDURL>/Layer3ForwardingSCPD.xml</SCPDURL>\n<controlURL>/upnp/control/Layer3Forwarding</controlURL>\n<eventSubURL>/upnp/event/Layer3Forwarding</eventSubURL>\n</service>\n</serviceList>\n<deviceList>\n<device>\n<deviceType>urn:schemas-upnp-org:device:WANDevice:1</deviceType>\n<friendlyName>WANDevice:1</friendlyName>\n<manufacturer>Cisco</manufacturer>\n<manufacturerURL>http://www.cisco.com/</manufacturerURL>\n<modelDescription>CGA4131COM</modelDescription>\n<modelName>CGA4131COM</modelName>\n<modelNumber>CGA4131COM</modelNumber>\n<modelURL>http://www.cisco.com</modelURL>\n<serialNumber></serialNumber>\n<UDN>uuid:11111111-2222-3333-4444-555555555556</UDN>\n<UPC>CGA4131COM</UPC>\n<serviceList>\n<service>\n<serviceType>urn:schemas-upnp-org:service:WANCommonInterfaceConfig:1</serviceType>\n<serviceId>urn:upnp-org:serviceId:WANCommonIFC1</serviceId>\n<SCPDURL>/WANCommonInterfaceConfigSCPD.xml</SCPDURL>\n<controlURL>/upnp/control/WANCommonInterfaceConfig0</controlURL>\n<eventSubURL>/upnp/event/WANCommonInterfaceConfig0</eventSubURL>\n</service>\n</serviceList>\n<deviceList>\n    <device>\n        <deviceType>urn:schemas-upnp-org:device:WANConnectionDevice:1</deviceType>\n        <friendlyName>WANConnectionDevice:1</friendlyName>\n        <manufacturer>Cisco</manufacturer>\n        <manufacturerURL>http://www.cisco.com/</manufacturerURL>\n        <modelDescription>CGA4131COM</modelDescription>\n        <modelName>CGA4131COM</modelName>\n        <modelNumber>CGA4131COM</modelNumber>\n        <modelURL>http://www.cisco.com</modelURL>\n        <serialNumber></serialNumber>\n        <UDN>uuid:11111111-2222-3333-4444-555555555555</UDN>\n        <UPC>CGA4131COM</UPC>\n        <serviceList>\n       <service>\n           <serviceType>urn:schemas-upnp-org:service:WANIPConnection:1</serviceType>\n           <serviceId>urn:upnp-org:serviceId:WANIPConn1</serviceId>\n           <SCPDURL>/WANIPConnectionServiceSCPD.xml</SCPDURL>\n           <controlURL>/upnp/control/WANIPConnection0</controlURL>\n           <eventSubURL>/upnp/event/WANIPConnection0</eventSubURL>\n       </service>\n        </serviceList>\n    </device>\n</deviceList>\n</device>\n</deviceList>\n<presentationURL>http://10.1.10.1/</presentationURL></device>\n</root>\n"

    response_bad_root_xmls = b"HTTP/1.1 200 OK\r\n" \
               b"CONTENT-LENGTH: 2972\r\n" \
               b"CONTENT-TYPE: text/xml\r\n" \
               b"DATE: Thu, 18 Oct 2018 01:20:23 GMT\r\n" \
               b"LAST-MODIFIED: Fri, 28 Sep 2018 18:35:48 GMT\r\n" \
               b"SERVER: Linux/3.14.28-Prod_17.2, UPnP/1.0, Portable SDK for UPnP devices/1.6.22\r\n" \
               b"X-User-Agent: redsonic\r\n" \
               b"CONNECTION: close\r\n" \
               b"\r\n" \
               b"<?xml version=\"1.0\"?>\n<root xmlns=\"urn:schemas-upnp--org:device-1-0\">\n<specVersion>\n<major>1</major>\n<minor>0</minor>\n</specVersion>\n<device>\n<deviceType>urn:schemas-upnp-org:device:InternetGatewayDevic3:1</deviceType>\n<friendlyName>CGA4131COM</friendlyName>\n<manufacturer>Cisco</manufacturer>\n<manufacturerURL>http://www.cisco.com/</manufacturerURL>\n<modelDescription>CGA4131COM</modelDescription>\n<modelName>CGA4131COM</modelName>\n<modelNumber>CGA4131COM</modelNumber>\n<modelURL>http://www.cisco.com</modelURL>\n<serialNumber></serialNumber>\n<UDN>uuid:11111111-2222-3333-4444-555555555556</UDN>\n<UPC>CGA4131COM</UPC>\n<serviceList>\n<service>\n<serviceType>urn:schemas-upnp-org:service:Layer3Forwarding:1</serviceType>\n<serviceId>urn:upnp-org:serviceId:L3Forwarding1</serviceId>\n<SCPDURL>/Layer3ForwardingSCPD.xml</SCPDURL>\n<controlURL>/upnp/control/Layer3Forwarding</controlURL>\n<eventSubURL>/upnp/event/Layer3Forwarding</eventSubURL>\n</service>\n</serviceList>\n<deviceList>\n<device>\n<deviceType>urn:schemas-upnp-org:device:WANDevice:1</deviceType>\n<friendlyName>WANDevice:1</friendlyName>\n<manufacturer>Cisco</manufacturer>\n<manufacturerURL>http://www.cisco.com/</manufacturerURL>\n<modelDescription>CGA4131COM</modelDescription>\n<modelName>CGA4131COM</modelName>\n<modelNumber>CGA4131COM</modelNumber>\n<modelURL>http://www.cisco.com</modelURL>\n<serialNumber></serialNumber>\n<UDN>uuid:11111111-2222-3333-4444-555555555556</UDN>\n<UPC>CGA4131COM</UPC>\n<serviceList>\n<service>\n<serviceType>urn:schemas-upnp-org:service:WANCommonInterfaceConfig:1</serviceType>\n<serviceId>urn:upnp-org:serviceId:WANCommonIFC1</serviceId>\n<SCPDURL>/WANCommonInterfaceConfigSCPD.xml</SCPDURL>\n<controlURL>/upnp/control/WANCommonInterfaceConfig0</controlURL>\n<eventSubURL>/upnp/event/WANCommonInterfaceConfig0</eventSubURL>\n</service>\n</serviceList>\n<deviceList>\n    <device>\n        <deviceType>urn:schemas-upnp-org:device:WANConnectionDevice:1</deviceType>\n        <friendlyName>WANConnectionDevice:1</friendlyName>\n        <manufacturer>Cisco</manufacturer>\n        <manufacturerURL>http://www.cisco.com/</manufacturerURL>\n        <modelDescription>CGA4131COM</modelDescription>\n        <modelName>CGA4131COM</modelName>\n        <modelNumber>CGA4131COM</modelNumber>\n        <modelURL>http://www.cisco.com</modelURL>\n        <serialNumber></serialNumber>\n        <UDN>uuid:11111111-2222-3333-4444-555555555555</UDN>\n        <UPC>CGA4131COM</UPC>\n        <serviceList>\n       <service>\n           <serviceType>urn:schemas-upnp-org:service:WANIPConnection:1</serviceType>\n           <serviceId>urn:upnp-org:serviceId:WANIPConn1</serviceId>\n           <SCPDURL>/WANIPConnectionServiceSCPD.xml</SCPDURL>\n           <controlURL>/upnp/control/WANIPConnection0</controlURL>\n           <eventSubURL>/upnp/event/WANIPConnection0</eventSubURL>\n       </service>\n        </serviceList>\n    </device>\n</deviceList>\n</device>\n</deviceList>\n<presentationURL>http://10.1.10.1/</presentationURL></device>\n</root>\n"

    expected_parsed = {
        'specVersion': {'major': '1', 'minor': '0'},
        'device': {
            'deviceType': 'urn:schemas-upnp-org:device:InternetGatewayDevice:1',
            'friendlyName': 'CGA4131COM',
            'manufacturer': 'Cisco',
            'manufacturerURL': 'http://www.cisco.com/',
            'modelDescription': 'CGA4131COM',
            'modelName': 'CGA4131COM',
            'modelNumber': 'CGA4131COM',
            'modelURL': 'http://www.cisco.com',
            'UDN': 'uuid:11111111-2222-3333-4444-555555555556',
            'UPC': 'CGA4131COM',
            'serviceList': {
                'service': {
                    'serviceType': 'urn:schemas-upnp-org:service:Layer3Forwarding:1',
                    'serviceId': 'urn:upnp-org:serviceId:L3Forwarding1',
                    'SCPDURL': '/Layer3ForwardingSCPD.xml',
                    'controlURL': '/upnp/control/Layer3Forwarding',
                    'eventSubURL': '/upnp/event/Layer3Forwarding'
                }
            },
            'deviceList': {
                'device': {
                    'deviceType': 'urn:schemas-upnp-org:device:WANDevice:1',
                    'friendlyName': 'WANDevice:1',
                    'manufacturer': 'Cisco',
                    'manufacturerURL': 'http://www.cisco.com/',
                    'modelDescription': 'CGA4131COM',
                    'modelName': 'CGA4131COM',
                    'modelNumber': 'CGA4131COM',
                    'modelURL': 'http://www.cisco.com',
                    'UDN': 'uuid:11111111-2222-3333-4444-555555555556',
                    'UPC': 'CGA4131COM',
                    'serviceList': {
                        'service': {
                            'serviceType': 'urn:schemas-upnp-org:service:WANCommonInterfaceConfig:1',
                            'serviceId': 'urn:upnp-org:serviceId:WANCommonIFC1',
                            'SCPDURL': '/WANCommonInterfaceConfigSCPD.xml',
                            'controlURL': '/upnp/control/WANCommonInterfaceConfig0',
                            'eventSubURL': '/upnp/event/WANCommonInterfaceConfig0'
                        }
                    },
                    'deviceList': {
                        'device': {
                            'deviceType': 'urn:schemas-upnp-org:device:WANConnectionDevice:1',
                            'friendlyName': 'WANConnectionDevice:1',
                            'manufacturer': 'Cisco',
                            'manufacturerURL': 'http://www.cisco.com/',
                            'modelDescription': 'CGA4131COM',
                            'modelName': 'CGA4131COM',
                            'modelNumber': 'CGA4131COM',
                            'modelURL': 'http://www.cisco.com',
                            'UDN': 'uuid:11111111-2222-3333-4444-555555555555',
                            'UPC': 'CGA4131COM',
                            'serviceList': {
                                'service': {
                                    'serviceType': 'urn:schemas-upnp-org:service:WANIPConnection:1',
                                    'serviceId': 'urn:upnp-org:serviceId:WANIPConn1',
                                    'SCPDURL': '/WANIPConnectionServiceSCPD.xml',
                                    'controlURL': '/upnp/control/WANIPConnection0',
                                    'eventSubURL': '/upnp/event/WANIPConnection0'
                                }
                            }
                        }
                    }
                }
            },
            'presentationURL': 'http://10.1.10.1/'
        }
    }

    def test_serialize_get(self):
        self.assertEqual(serialize_scpd_get(self.path, self.lan_address), self.get_request)
        self.assertEqual(serialize_scpd_get(self.path, 'http://' + self.lan_address), self.get_request)
        self.assertEqual(serialize_scpd_get(self.path, 'http://' + self.lan_address + ':1337'), self.get_request)
        self.assertEqual(serialize_scpd_get(self.path, self.lan_address + ':1337'), self.get_request)

    def test_parse_device_response_xml(self):
        self.assertDictEqual(
            xml_to_dict('<?xml version="1.0"?>\n<root xmlns="urn:schemas-upnp-org:device-1-0">\n\t<specVersion>\n\t\t<major>1</major>\n\t\t<minor>0</minor>\n\t</specVersion>\n\t<URLBase>http://10.0.0.1:49152</URLBase>\n\t<device>\n\t\t<deviceType>urn:schemas-upnp-org:device:InternetGatewayDevice:1</deviceType>\n\t\t<friendlyName>Wireless Broadband Router</friendlyName>\n\t\t<manufacturer>D-Link Corporation</manufacturer>\n\t\t<manufacturerURL>http://www.dlink.com</manufacturerURL>\n\t\t<modelDescription>D-Link Router</modelDescription>\n\t\t<modelName>D-Link Router</modelName>\n\t\t<modelNumber>DIR-890L</modelNumber>\n\t\t<modelURL>http://www.dlink.com</modelURL>\n\t\t<serialNumber>120</serialNumber>\n\t\t<UDN>uuid:11111111-2222-3333-4444-555555555555</UDN>\n\t\t<iconList>\n\t\t\t<icon>\n\t\t\t\t<mimetype>image/gif</mimetype>\n\t\t\t\t<width>118</width>\n\t\t\t\t<height>119</height>\n\t\t\t\t<depth>8</depth>\n\t\t\t\t<url>/ligd.gif</url>\n\t\t\t</icon>\n\t\t</iconList>\n\t\t<serviceList>\n\t\t\t<service>\n\t\t\t\t<serviceType>urn:schemas-microsoft-com:service:OSInfo:1</serviceType>\n\t\t\t\t<serviceId>urn:microsoft-com:serviceId:OSInfo1</serviceId>\n\t\t\t\t<controlURL>/soap.cgi?service=OSInfo1</controlURL>\n\t\t\t\t<eventSubURL>/gena.cgi?service=OSInfo1</eventSubURL>\n\t\t\t\t<SCPDURL>/OSInfo.xml</SCPDURL>\n\t\t\t</service>\n\t\t\t<service>\n\t\t\t\t<serviceType>urn:schemas-upnp-org:service:Layer3Forwarding:1</serviceType>\n\t\t\t\t<serviceId>urn:upnp-org:serviceId:L3Forwarding1</serviceId>\n\t\t\t\t<controlURL>/soap.cgi?service=L3Forwarding1</controlURL>\n\t\t\t\t<eventSubURL>/gena.cgi?service=L3Forwarding1</eventSubURL>\n\t\t\t\t<SCPDURL>/Layer3Forwarding.xml</SCPDURL>\n\t\t\t</service>\n\t\t</serviceList>\n\t\t<deviceList>\n\t\t\t<device>\n\t\t\t\t<deviceType>urn:schemas-upnp-org:device:WANDevice:1</deviceType>\n\t\t\t\t<friendlyName>WANDevice</friendlyName>\n\t\t\t\t<manufacturer>D-Link</manufacturer>\n\t\t\t\t<manufacturerURL>http://www.dlink.com</manufacturerURL>\n\t\t\t\t<modelDescription>WANDevice</modelDescription>\n\t\t\t\t<modelName>DIR-890L</modelName>\n\t\t\t\t<modelNumber>1</modelNumber>\n\t\t\t\t<modelURL>http://www.dlink.com</modelURL>\n\t\t\t\t<serialNumber>120</serialNumber>\n\t\t\t\t<UDN>uuid:11111111-2222-3333-4444-555555555555</UDN>\n\t\t\t\t<serviceList>\n\t\t\t\t\t<service>\n\t\t\t\t\t\t<serviceType>urn:schemas-upnp-org:service:WANCommonInterfaceConfig:1</serviceType>\n\t\t\t\t\t\t<serviceId>urn:upnp-org:serviceId:WANCommonIFC1</serviceId>\n\t\t\t\t\t\t<controlURL>/soap.cgi?service=WANCommonIFC1</controlURL>\n\t\t\t\t\t\t<eventSubURL>/gena.cgi?service=WANCommonIFC1</eventSubURL>\n\t\t\t\t\t\t<SCPDURL>/WANCommonInterfaceConfig.xml</SCPDURL>\n\t\t\t\t\t</service>\n\t\t\t\t</serviceList>\n\t\t\t\t<deviceList>\n\t\t\t\t\t<device>\n\t\t\t\t\t\t<deviceType>urn:schemas-upnp-org:device:WANConnectionDevice:1</deviceType>\n\t\t\t\t\t\t<friendlyName>WANConnectionDevice</friendlyName>\n\t\t\t\t\t\t<manufacturer>D-Link</manufacturer>\n\t\t\t\t\t\t<manufacturerURL>http://www.dlink.com</manufacturerURL>\n\t\t\t\t\t\t<modelDescription>WanConnectionDevice</modelDescription>\n\t\t\t\t\t\t<modelName>DIR-890L</modelName>\n\t\t\t\t\t\t<modelNumber>1</modelNumber>\n\t\t\t\t\t\t<modelURL>http://www.dlink.com</modelURL>\n\t\t\t\t\t\t<serialNumber>120</serialNumber>\n\t\t\t\t\t\t<UDN>uuid:11111111-2222-3333-4444-555555555555</UDN>\n\t\t\t\t\t\t<serviceList>\n\t\t\t\t\t\t\t<service>\n\t\t\t\t\t\t\t\t<serviceType>urn:schemas-upnp-org:service:WANEthernetLinkConfig:1</serviceType>\n\t\t\t\t\t\t\t\t<serviceId>urn:upnp-org:serviceId:WANEthLinkC1</serviceId>\n\t\t\t\t\t\t\t\t<controlURL>/soap.cgi?service=WANEthLinkC1</controlURL>\n\t\t\t\t\t\t\t\t<eventSubURL>/gena.cgi?service=WANEthLinkC1</eventSubURL>\n\t\t\t\t\t\t\t\t<SCPDURL>/WANEthernetLinkConfig.xml</SCPDURL>\n\t\t\t\t\t\t\t</service>\n\t\t\t\t\t\t\t<service>\n\t\t\t\t\t\t\t\t<serviceType>urn:schemas-upnp-org:service:WANIPConnection:1</serviceType>\n\t\t\t\t\t\t\t\t<serviceId>urn:upnp-org:serviceId:WANIPConn1</serviceId>\n\t\t\t\t\t\t\t\t<controlURL>/soap.cgi?service=WANIPConn1</controlURL>\n\t\t\t\t\t\t\t\t<eventSubURL>/gena.cgi?service=WANIPConn1</eventSubURL>\n\t\t\t\t\t\t\t\t<SCPDURL>/WANIPConnection.xml</SCPDURL>\n\t\t\t\t\t\t\t</service>\n\t\t\t\t\t\t</serviceList>\n\t\t\t\t\t</device>\n\t\t\t\t</deviceList>\n\t\t\t</device>\n\t\t</deviceList>\n\t\t<presentationURL>http://10.0.0.1</presentationURL>\n\t</device>\n</root>\n'),
            {'{urn:schemas-upnp-org:device-1-0}root': {
                '{urn:schemas-upnp-org:device-1-0}specVersion': {'{urn:schemas-upnp-org:device-1-0}major': '1',
                                                                 '{urn:schemas-upnp-org:device-1-0}minor': '0'},
                '{urn:schemas-upnp-org:device-1-0}URLBase': 'http://10.0.0.1:49152',
                '{urn:schemas-upnp-org:device-1-0}device': {
                    '{urn:schemas-upnp-org:device-1-0}deviceType': 'urn:schemas-upnp-org:device:InternetGatewayDevice:1',
                    '{urn:schemas-upnp-org:device-1-0}friendlyName': 'Wireless Broadband Router',
                    '{urn:schemas-upnp-org:device-1-0}manufacturer': 'D-Link Corporation',
                    '{urn:schemas-upnp-org:device-1-0}manufacturerURL': 'http://www.dlink.com',
                    '{urn:schemas-upnp-org:device-1-0}modelDescription': 'D-Link Router',
                    '{urn:schemas-upnp-org:device-1-0}modelName': 'D-Link Router',
                    '{urn:schemas-upnp-org:device-1-0}modelNumber': 'DIR-890L',
                    '{urn:schemas-upnp-org:device-1-0}modelURL': 'http://www.dlink.com',
                    '{urn:schemas-upnp-org:device-1-0}serialNumber': '120',
                    '{urn:schemas-upnp-org:device-1-0}UDN': 'uuid:11111111-2222-3333-4444-555555555555',
                    '{urn:schemas-upnp-org:device-1-0}iconList': {'{urn:schemas-upnp-org:device-1-0}icon': {
                        '{urn:schemas-upnp-org:device-1-0}mimetype': 'image/gif',
                        '{urn:schemas-upnp-org:device-1-0}width': '118',
                        '{urn:schemas-upnp-org:device-1-0}height': '119', '{urn:schemas-upnp-org:device-1-0}depth': '8',
                        '{urn:schemas-upnp-org:device-1-0}url': '/ligd.gif'}},
                    '{urn:schemas-upnp-org:device-1-0}serviceList': {'{urn:schemas-upnp-org:device-1-0}service': [
                        {'{urn:schemas-upnp-org:device-1-0}serviceType': 'urn:schemas-microsoft-com:service:OSInfo:1',
                         '{urn:schemas-upnp-org:device-1-0}serviceId': 'urn:microsoft-com:serviceId:OSInfo1',
                         '{urn:schemas-upnp-org:device-1-0}controlURL': '/soap.cgi?service=OSInfo1',
                         '{urn:schemas-upnp-org:device-1-0}eventSubURL': '/gena.cgi?service=OSInfo1',
                         '{urn:schemas-upnp-org:device-1-0}SCPDURL': '/OSInfo.xml'}, {
                            '{urn:schemas-upnp-org:device-1-0}serviceType': 'urn:schemas-upnp-org:service:Layer3Forwarding:1',
                            '{urn:schemas-upnp-org:device-1-0}serviceId': 'urn:upnp-org:serviceId:L3Forwarding1',
                            '{urn:schemas-upnp-org:device-1-0}controlURL': '/soap.cgi?service=L3Forwarding1',
                            '{urn:schemas-upnp-org:device-1-0}eventSubURL': '/gena.cgi?service=L3Forwarding1',
                            '{urn:schemas-upnp-org:device-1-0}SCPDURL': '/Layer3Forwarding.xml'}]},
                    '{urn:schemas-upnp-org:device-1-0}deviceList': {'{urn:schemas-upnp-org:device-1-0}device': {
                        '{urn:schemas-upnp-org:device-1-0}deviceType': 'urn:schemas-upnp-org:device:WANDevice:1',
                        '{urn:schemas-upnp-org:device-1-0}friendlyName': 'WANDevice',
                        '{urn:schemas-upnp-org:device-1-0}manufacturer': 'D-Link',
                        '{urn:schemas-upnp-org:device-1-0}manufacturerURL': 'http://www.dlink.com',
                        '{urn:schemas-upnp-org:device-1-0}modelDescription': 'WANDevice',
                        '{urn:schemas-upnp-org:device-1-0}modelName': 'DIR-890L',
                        '{urn:schemas-upnp-org:device-1-0}modelNumber': '1',
                        '{urn:schemas-upnp-org:device-1-0}modelURL': 'http://www.dlink.com',
                        '{urn:schemas-upnp-org:device-1-0}serialNumber': '120',
                        '{urn:schemas-upnp-org:device-1-0}UDN': 'uuid:11111111-2222-3333-4444-555555555555',
                        '{urn:schemas-upnp-org:device-1-0}serviceList': {'{urn:schemas-upnp-org:device-1-0}service': {
                            '{urn:schemas-upnp-org:device-1-0}serviceType': 'urn:schemas-upnp-org:service:WANCommonInterfaceConfig:1',
                            '{urn:schemas-upnp-org:device-1-0}serviceId': 'urn:upnp-org:serviceId:WANCommonIFC1',
                            '{urn:schemas-upnp-org:device-1-0}controlURL': '/soap.cgi?service=WANCommonIFC1',
                            '{urn:schemas-upnp-org:device-1-0}eventSubURL': '/gena.cgi?service=WANCommonIFC1',
                            '{urn:schemas-upnp-org:device-1-0}SCPDURL': '/WANCommonInterfaceConfig.xml'}},
                        '{urn:schemas-upnp-org:device-1-0}deviceList': {'{urn:schemas-upnp-org:device-1-0}device': {
                            '{urn:schemas-upnp-org:device-1-0}deviceType': 'urn:schemas-upnp-org:device:WANConnectionDevice:1',
                            '{urn:schemas-upnp-org:device-1-0}friendlyName': 'WANConnectionDevice',
                            '{urn:schemas-upnp-org:device-1-0}manufacturer': 'D-Link',
                            '{urn:schemas-upnp-org:device-1-0}manufacturerURL': 'http://www.dlink.com',
                            '{urn:schemas-upnp-org:device-1-0}modelDescription': 'WanConnectionDevice',
                            '{urn:schemas-upnp-org:device-1-0}modelName': 'DIR-890L',
                            '{urn:schemas-upnp-org:device-1-0}modelNumber': '1',
                            '{urn:schemas-upnp-org:device-1-0}modelURL': 'http://www.dlink.com',
                            '{urn:schemas-upnp-org:device-1-0}serialNumber': '120',
                            '{urn:schemas-upnp-org:device-1-0}UDN': 'uuid:11111111-2222-3333-4444-555555555555',
                            '{urn:schemas-upnp-org:device-1-0}serviceList': {
                                '{urn:schemas-upnp-org:device-1-0}service': [{
                                                                                 '{urn:schemas-upnp-org:device-1-0}serviceType': 'urn:schemas-upnp-org:service:WANEthernetLinkConfig:1',
                                                                                 '{urn:schemas-upnp-org:device-1-0}serviceId': 'urn:upnp-org:serviceId:WANEthLinkC1',
                                                                                 '{urn:schemas-upnp-org:device-1-0}controlURL': '/soap.cgi?service=WANEthLinkC1',
                                                                                 '{urn:schemas-upnp-org:device-1-0}eventSubURL': '/gena.cgi?service=WANEthLinkC1',
                                                                                 '{urn:schemas-upnp-org:device-1-0}SCPDURL': '/WANEthernetLinkConfig.xml'},
                                                                             {
                                                                                 '{urn:schemas-upnp-org:device-1-0}serviceType': 'urn:schemas-upnp-org:service:WANIPConnection:1',
                                                                                 '{urn:schemas-upnp-org:device-1-0}serviceId': 'urn:upnp-org:serviceId:WANIPConn1',
                                                                                 '{urn:schemas-upnp-org:device-1-0}controlURL': '/soap.cgi?service=WANIPConn1',
                                                                                 '{urn:schemas-upnp-org:device-1-0}eventSubURL': '/gena.cgi?service=WANIPConn1',
                                                                                 '{urn:schemas-upnp-org:device-1-0}SCPDURL': '/WANIPConnection.xml'}]}}}}},
                    '{urn:schemas-upnp-org:device-1-0}presentationURL': 'http://10.0.0.1'}}}
        )

    def test_deserialize_get_response(self):
        self.assertDictEqual(deserialize_scpd_get_response(self.response), self.expected_parsed)

    def test_deserialize_blank(self):
        self.assertDictEqual(deserialize_scpd_get_response(b''), {})

    def test_fail_to_deserialize_invalid_root_device(self):
        with self.assertRaises(UPnPError):
            deserialize_scpd_get_response(self.response_bad_root_device_name)

    def test_fail_to_deserialize_invalid_root_xmls(self):
        with self.assertRaises(UPnPError):
            deserialize_scpd_get_response(self.response_bad_root_xmls)

    def test_deserialize_to_device_object(self):
        devices = []
        services = []
        device = Device(devices, services, **get_dict_val_case_insensitive(self.expected_parsed, "device"))
        expected_result = {
            'deviceType': 'urn:schemas-upnp-org:device:InternetGatewayDevice:1',
            'friendlyName': 'CGA4131COM',
            'manufacturer': 'Cisco',
            'manufacturerURL': 'http://www.cisco.com/',
            'modelDescription': 'CGA4131COM',
            'modelName': 'CGA4131COM',
            'modelNumber': 'CGA4131COM',
            'modelURL': 'http://www.cisco.com',
            'udn': 'uuid:11111111-2222-3333-4444-555555555556',
            'upc': 'CGA4131COM',
            'serviceList': {
                'service': {
                    'serviceType': 'urn:schemas-upnp-org:service:Layer3Forwarding:1',
                    'serviceId': 'urn:upnp-org:serviceId:L3Forwarding1',
                    'SCPDURL': '/Layer3ForwardingSCPD.xml',
                    'controlURL': '/upnp/control/Layer3Forwarding',
                    'eventSubURL': '/upnp/event/Layer3Forwarding'
                }
            },
            'deviceList': {
                'device': {
                    'deviceType': 'urn:schemas-upnp-org:device:WANDevice:1',
                    'friendlyName': 'WANDevice:1',
                    'manufacturer': 'Cisco',
                    'manufacturerURL': 'http://www.cisco.com/',
                    'modelDescription': 'CGA4131COM',
                    'modelName': 'CGA4131COM',
                    'modelNumber': 'CGA4131COM',
                    'modelURL': 'http://www.cisco.com',
                    'UDN': 'uuid:11111111-2222-3333-4444-555555555556',
                    'UPC': 'CGA4131COM',
                    'serviceList': {
                        'service': {
                            'serviceType': 'urn:schemas-upnp-org:service:WANCommonInterfaceConfig:1',
                            'serviceId': 'urn:upnp-org:serviceId:WANCommonIFC1',
                            'SCPDURL': '/WANCommonInterfaceConfigSCPD.xml',
                            'controlURL': '/upnp/control/WANCommonInterfaceConfig0',
                            'eventSubURL': '/upnp/event/WANCommonInterfaceConfig0'
                        }
                    },
                    'deviceList': {
                        'device': {
                            'deviceType': 'urn:schemas-upnp-org:device:WANConnectionDevice:1',
                            'friendlyName': 'WANConnectionDevice:1',
                            'manufacturer': 'Cisco',
                            'manufacturerURL': 'http://www.cisco.com/',
                            'modelDescription': 'CGA4131COM',
                            'modelName': 'CGA4131COM',
                            'modelNumber': 'CGA4131COM',
                            'modelURL': 'http://www.cisco.com',
                            'UDN': 'uuid:11111111-2222-3333-4444-555555555555',
                            'UPC': 'CGA4131COM',
                            'serviceList': {
                                'service': {
                                    'serviceType': 'urn:schemas-upnp-org:service:WANIPConnection:1',
                                    'serviceId': 'urn:upnp-org:serviceId:WANIPConn1',
                                    'SCPDURL': '/WANIPConnectionServiceSCPD.xml',
                                    'controlURL': '/upnp/control/WANIPConnection0',
                                    'eventSubURL': '/upnp/event/WANIPConnection0'
                                }
                            }
                        }
                    }
                }
            }, 'presentationURL': 'http://10.1.10.1/'
        }
        self.assertDictEqual(expected_result, device.as_dict())

    def test_deserialize_another_device(self):
        xml_bytes = b"<?xml version=\"1.0\"?>\n<root xmlns=\"urn:schemas-upnp-org:device-1-0\">\n<specVersion>\n<major>1</major>\n<minor>0</minor>\n</specVersion>\n<device>\n<deviceType>urn:schemas-upnp-org:device:InternetGatewayDevice:1</deviceType>\n<friendlyName>CGA4131COM</friendlyName>\n<manufacturer>Cisco</manufacturer>\n<manufacturerURL>http://www.cisco.com/</manufacturerURL>\n<modelDescription>CGA4131COM</modelDescription>\n<modelName>CGA4131COM</modelName>\n<modelNumber>CGA4131COM</modelNumber>\n<modelURL>http://www.cisco.com</modelURL>\n<serialNumber></serialNumber>\n<UDN>uuid:11111111-2222-3333-4444-555555555556</UDN>\n<UPC>CGA4131COM</UPC>\n<serviceList>\n<service>\n<serviceType>urn:schemas-upnp-org:service:Layer3Forwarding:1</serviceType>\n<serviceId>urn:upnp-org:serviceId:L3Forwarding1</serviceId>\n<SCPDURL>/Layer3ForwardingSCPD.xml</SCPDURL>\n<controlURL>/upnp/control/Layer3Forwarding</controlURL>\n<eventSubURL>/upnp/event/Layer3Forwarding</eventSubURL>\n</service>\n</serviceList>\n<deviceList>\n<device>\n<deviceType>urn:schemas-upnp-org:device:WANDevice:1</deviceType>\n<friendlyName>WANDevice:1</friendlyName>\n<manufacturer>Cisco</manufacturer>\n<manufacturerURL>http://www.cisco.com/</manufacturerURL>\n<modelDescription>CGA4131COM</modelDescription>\n<modelName>CGA4131COM</modelName>\n<modelNumber>CGA4131COM</modelNumber>\n<modelURL>http://www.cisco.com</modelURL>\n<serialNumber></serialNumber>\n<UDN>uuid:ebf5a0a0-1dd1-11b2-a92f-603d266f9915</UDN>\n<UPC>CGA4131COM</UPC>\n<serviceList>\n<service>\n<serviceType>urn:schemas-upnp-org:service:WANCommonInterfaceConfig:1</serviceType>\n<serviceId>urn:upnp-org:serviceId:WANCommonIFC1</serviceId>\n<SCPDURL>/WANCommonInterfaceConfigSCPD.xml</SCPDURL>\n<controlURL>/upnp/control/WANCommonInterfaceConfig0</controlURL>\n<eventSubURL>/upnp/event/WANCommonInterfaceConfig0</eventSubURL>\n</service>\n</serviceList>\n<deviceList>\n    <device>\n        <deviceType>urn:schemas-upnp-org:device:WANConnectionDevice:1</deviceType>\n        <friendlyName>WANConnectionDevice:1</friendlyName>\n        <manufacturer>Cisco</manufacturer>\n        <manufacturerURL>http://www.cisco.com/</manufacturerURL>\n        <modelDescription>CGA4131COM</modelDescription>\n        <modelName>CGA4131COM</modelName>\n        <modelNumber>CGA4131COM</modelNumber>\n        <modelURL>http://www.cisco.com</modelURL>\n        <serialNumber></serialNumber>\n        <UDN>uuid:11111111-2222-3333-4444-555555555555</UDN>\n        <UPC>CGA4131COM</UPC>\n        <serviceList>\n       <service>\n           <serviceType>urn:schemas-upnp-org:service:WANIPConnection:1</serviceType>\n           <serviceId>urn:upnp-org:serviceId:WANIPConn1</serviceId>\n           <SCPDURL>/WANIPConnectionServiceSCPD.xml</SCPDURL>\n           <controlURL>/upnp/control/WANIPConnection0</controlURL>\n           <eventSubURL>/upnp/event/WANIPConnection0</eventSubURL>\n       </service>\n        </serviceList>\n    </device>\n</deviceList>\n</device>\n</deviceList>\n<presentationURL>http://10.1.10.1/</presentationURL></device>\n</root>\n"
        expected_parsed = {
            'specVersion': {'major': '1', 'minor': '0'},
            'device': {
                'deviceType': 'urn:schemas-upnp-org:device:InternetGatewayDevice:1',
                'friendlyName': 'CGA4131COM',
                'manufacturer': 'Cisco',
                'manufacturerURL': 'http://www.cisco.com/',
                'modelDescription': 'CGA4131COM',
                'modelName': 'CGA4131COM',
                'modelNumber': 'CGA4131COM',
                'modelURL': 'http://www.cisco.com',
                'UDN': 'uuid:11111111-2222-3333-4444-555555555556',
                'UPC': 'CGA4131COM',
                'serviceList': {
                    'service': {
                        'serviceType': 'urn:schemas-upnp-org:service:Layer3Forwarding:1',
                        'serviceId': 'urn:upnp-org:serviceId:L3Forwarding1',
                        'SCPDURL': '/Layer3ForwardingSCPD.xml',
                        'controlURL': '/upnp/control/Layer3Forwarding',
                        'eventSubURL': '/upnp/event/Layer3Forwarding'
                    }
                },
                'deviceList': {
                    'device': {
                        'deviceType': 'urn:schemas-upnp-org:device:WANDevice:1',
                        'friendlyName': 'WANDevice:1',
                        'manufacturer': 'Cisco',
                        'manufacturerURL': 'http://www.cisco.com/',
                        'modelDescription': 'CGA4131COM',
                        'modelName': 'CGA4131COM',
                        'modelNumber': 'CGA4131COM',
                        'modelURL': 'http://www.cisco.com',
                        'UDN': 'uuid:ebf5a0a0-1dd1-11b2-a92f-603d266f9915',
                        'UPC': 'CGA4131COM',
                        'serviceList': {
                            'service': {
                                'serviceType': 'urn:schemas-upnp-org:service:WANCommonInterfaceConfig:1',
                                'serviceId': 'urn:upnp-org:serviceId:WANCommonIFC1',
                                'SCPDURL': '/WANCommonInterfaceConfigSCPD.xml',
                                'controlURL': '/upnp/control/WANCommonInterfaceConfig0',
                                'eventSubURL': '/upnp/event/WANCommonInterfaceConfig0'
                            }
                        },
                        'deviceList': {
                            'device': {
                                'deviceType': 'urn:schemas-upnp-org:device:WANConnectionDevice:1',
                                'friendlyName': 'WANConnectionDevice:1',
                                'manufacturer': 'Cisco',
                                'manufacturerURL': 'http://www.cisco.com/',
                                'modelDescription': 'CGA4131COM',
                                'modelName': 'CGA4131COM',
                                'modelNumber': 'CGA4131COM',
                                'modelURL': 'http://www.cisco.com',
                                'UDN': 'uuid:11111111-2222-3333-4444-555555555555',
                                'UPC': 'CGA4131COM',
                                'serviceList': {
                                    'service': {
                                        'serviceType': 'urn:schemas-upnp-org:service:WANIPConnection:1',
                                        'serviceId': 'urn:upnp-org:serviceId:WANIPConn1',
                                        'SCPDURL': '/WANIPConnectionServiceSCPD.xml',
                                        'controlURL': '/upnp/control/WANIPConnection0',
                                        'eventSubURL': '/upnp/event/WANIPConnection0'
                                    }
                                }
                            }
                        }
                    }
                },
                'presentationURL': 'http://10.1.10.1/'
            }
        }
        self.assertDictEqual(expected_parsed, deserialize_scpd_get_response(xml_bytes))



import re
import logging
import binascii
import json
from collections import OrderedDict
from typing import List, Optional, Dict, Union, Tuple, Callable
from aioupnp.fault import UPnPError
from aioupnp.constants import line_separator

log = logging.getLogger(__name__)

_template = "(?i)^(%s):[ ]*(.*)$"
vendor_pattern = re.compile("^([\w|\d]*)\.([\w|\d]*\.com):([ \"|\w|\d\:]*)$")


def match_vendor(line: str) -> Optional[Tuple[str, str]]:
    match: List[Tuple[str, str, str]] = vendor_pattern.findall(line)
    if match:
        vendor_key: str = match[-1][0].lstrip(" ").rstrip(" ")
        vendor_value: str = match[-1][2].lstrip(" ").rstrip(" ")
        return vendor_key, vendor_value
    return None


def compile_find(pattern: str) -> Callable[[str], Optional[str]]:
    p = re.compile(pattern)

    def find(line: str) -> Optional[str]:
        result: List[List[str]] = []
        for outer in p.findall(line):
            result.append([])
            for inner in outer:
                result[-1].append(inner)
        if result:
            return result[-1][-1].lstrip(" ").rstrip(" ")
        return None

    return find


ssdp_datagram_patterns: Dict[str, Callable[[str], Optional[str]]] = {
    'host': compile_find("(?i)^(host):(.*)$"),
    'st': compile_find(_template % 'st'),
    'man': compile_find(_template % 'man'),
    'mx': compile_find(_template % 'mx'),
    'nt': compile_find(_template % 'nt'),
    'nts': compile_find(_template % 'nts'),
    'usn': compile_find(_template % 'usn'),
    'location': compile_find(_template % 'location'),
    'cache_control': compile_find(_template % 'cache[-|_]control'),
    'server': compile_find(_template % 'server'),
}


class SSDPDatagram:
    _M_SEARCH = "M-SEARCH"
    _NOTIFY = "NOTIFY"
    _OK = "OK"

    _start_lines: Dict[str, str] = {
        _M_SEARCH: "M-SEARCH * HTTP/1.1",
        _NOTIFY: "NOTIFY * HTTP/1.1",
        _OK: "HTTP/1.1 200 OK"
    }

    _friendly_names: Dict[str, str] = {
        _M_SEARCH: "m-search",
        _NOTIFY: "notify",
        _OK: "m-search response"
    }

    _vendor_field_pattern = vendor_pattern

    _required_fields: Dict[str, List[str]] = {
        _M_SEARCH: [
            'host',
            'man',
            'mx',
            'st',
        ],
        _NOTIFY: [
            'host',
            'location',
            'nt',
            'nts',
            'server',
            'usn',
        ],
        _OK: [
            'cache_control',
            # 'date',
            # 'ext',
            'location',
            'server',
            'st',
            'usn'
        ]
    }

    def __init__(self, packet_type: str, kwargs: Optional[Dict[str, Union[str, int]]] = None) -> None:
        if packet_type not in [self._M_SEARCH, self._NOTIFY, self._OK]:
            raise UPnPError("unknown packet type: {}".format(packet_type))
        self._packet_type = packet_type
        kw: Dict[str, Union[str, int]] = kwargs or OrderedDict()
        self._field_order: List[str] = [
            k.lower().replace("-", "_") for k in kw.keys()
        ]
        self.host: Optional[str] = None
        self.man: Optional[str] = None
        self.mx: Optional[Union[str, int]] = None
        self.st: Optional[str] = None
        self.nt: Optional[str] = None
        self.nts: Optional[str] = None
        self.usn: Optional[str] = None
        self.location: Optional[str] = None
        self.cache_control: Optional[str] = None
        self.server: Optional[str] = None
        self.date: Optional[str] = None
        self.ext: Optional[str] = None
        for k, v in kw.items():
            normalized = k.lower().replace("-", "_")
            if not normalized.startswith("_") and hasattr(self, normalized):
                if getattr(self, normalized, None) is None:
                    setattr(self, normalized, v)
        self._case_mappings: Dict[str, str] = {k.lower(): k for k in kw.keys()}
        for k in self._required_fields[self._packet_type]:
            if getattr(self, k, None) is None:
                raise UPnPError("missing required field %s" % k)

    def get_cli_igd_kwargs(self) -> str:
        fields = []
        for field in self._field_order:
            v = getattr(self, field, None)
            if v is None:
                raise UPnPError("missing required field %s" % field)
            fields.append("--%s=%s" % (self._case_mappings.get(field, field), v))
        return " ".join(fields)

    def __repr__(self) -> str:
        return self.as_json()

    def __getitem__(self, item: str) -> Union[str, int]:
        for i in self._required_fields[self._packet_type]:
            if i.lower() == item.lower():
                return getattr(self, i)
        raise KeyError(item)

    def encode(self, trailing_newlines: int = 2) -> str:
        lines = [self._start_lines[self._packet_type]]
        lines.extend(
            f"{self._case_mappings.get(attr_name.lower(), attr_name.upper())}: {str(getattr(self, attr_name))}"
            for attr_name in self._field_order if attr_name in self._required_fields[self._packet_type]
        )
        serialized = line_separator.join(lines)
        for _ in range(trailing_newlines):
            serialized += line_separator
        return serialized

    def as_dict(self) -> Dict[str, Union[str, int]]:
        return self._lines_to_content_dict(self.encode().split(line_separator))

    def as_json(self) -> str:
        return json.dumps(self.as_dict(), indent=2)

    @classmethod
    def decode(cls, datagram: bytes) -> 'SSDPDatagram':
        try:
            packet = cls._from_string(datagram.decode())
        except UnicodeDecodeError:
            raise UPnPError(
                f"failed to decode datagram: {binascii.hexlify(datagram).decode()}"
            )
        if packet is None:
            raise UPnPError(
                f"failed to decode datagram: {binascii.hexlify(datagram).decode()}"
            )
        for attr_name in packet._required_fields[packet._packet_type]:
            if getattr(packet, attr_name, None) is None:
                raise UPnPError(
                    "required field for {} is missing from m-search response: {}".format(packet._packet_type, attr_name)
                )
        return packet

    @classmethod
    def _lines_to_content_dict(cls, lines: List[str]) -> Dict[str, Union[str, int]]:
        result: Dict[str, Union[str, int]] = OrderedDict()
        matched_keys: List[str] = []
        for line in lines:
            if not line:
                continue
            matched = False
            for name, pattern in ssdp_datagram_patterns.items():
                if name not in matched_keys:
                    if name.lower() == 'mx':
                        _matched_int = pattern(line)
                        if _matched_int is not None:
                            match_int = int(_matched_int)
                            result[line[:len(name)]] = match_int
                            matched = True
                            matched_keys.append(name)
                            break
                    else:
                        match = pattern(line)
                        if match is not None:
                            result[line[:len(name)]] = match
                            matched = True
                            matched_keys.append(name)
                            break
            if not matched:
                matched_vendor = match_vendor(line)
                if matched_vendor and matched_vendor[0] not in result:
                    result[matched_vendor[0]] = matched_vendor[1]
        return result

    @classmethod
    def _from_string(cls, datagram: str) -> Optional['SSDPDatagram']:
        lines = [l for l in datagram.split(line_separator) if l]
        if not lines:
            return None
        if lines[0] == cls._start_lines[cls._M_SEARCH]:
            return cls._from_request(lines[1:])
        if lines[0] in [cls._start_lines[cls._NOTIFY], cls._start_lines[cls._NOTIFY] + " "]:
            return cls._from_notify(lines[1:])
        if lines[0] == cls._start_lines[cls._OK]:
            return cls._from_response(lines[1:])
        return None

    @classmethod
    def _from_response(cls, lines: List) -> 'SSDPDatagram':
        return cls(cls._OK, cls._lines_to_content_dict(lines))

    @classmethod
    def _from_notify(cls, lines: List) -> 'SSDPDatagram':
        return cls(cls._NOTIFY, cls._lines_to_content_dict(lines))

    @classmethod
    def _from_request(cls, lines: List) -> 'SSDPDatagram':
        return cls(cls._M_SEARCH, cls._lines_to_content_dict(lines))



import socket
from collections import OrderedDict
import typing
import netifaces
from aioupnp.fault import UPnPError


def get_netifaces():
    return netifaces


def ifaddresses(iface: str) -> typing.Dict[int, typing.List[typing.Dict[str, str]]]:
    return get_netifaces().ifaddresses(iface)


def _get_interfaces() -> typing.List[str]:
    return get_netifaces().interfaces()


def _get_gateways() -> typing.Dict[typing.Union[str, int],
                                             typing.Union[typing.Dict[int, typing.Tuple[str, str]],
                                                          typing.List[typing.Tuple[str, str, bool]]]]:
    return get_netifaces().gateways()


def get_interfaces() -> typing.Dict[str, typing.Tuple[str, str]]:
    gateways = _get_gateways()
    infos = gateways[socket.AF_INET]
    assert isinstance(infos, list), TypeError(f"expected list from netifaces, got a dict")
    interface_infos: typing.List[typing.Tuple[str, str, bool]] = infos
    result: typing.Dict[str, typing.Tuple[str, str]] = OrderedDict(
        (interface_name, (router_address, ifaddresses(interface_name)[socket.AF_INET][0]['addr']))
        for router_address, interface_name, _ in interface_infos
    )
    for interface_name in _get_interfaces():
        if interface_name in ['lo', 'localhost'] or interface_name in result:
            continue
        addresses = ifaddresses(interface_name)
        if netifaces.AF_INET in addresses:
            address = addresses[netifaces.AF_INET][0]['addr']
            gateway_guess = ".".join(address.split(".")[:-1] + ["1"])
            result[interface_name] = (gateway_guess, address)
    _default = gateways['default']
    assert isinstance(_default, dict), TypeError(f"expected dict from netifaces, got a list")
    default: typing.Dict[int, typing.Tuple[str, str]] = _default
    result['default'] = result[default[socket.AF_INET][1]]
    return result


def get_gateway_and_lan_addresses(interface_name: str) -> typing.Tuple[str, str]:
    for iface_name, (gateway, lan) in get_interfaces().items():
        if interface_name == iface_name:
            return gateway, lan
    raise UPnPError(f'failed to get lan and gateway addresses for {interface_name}')



import typing
from collections import OrderedDict


str_any_dict = typing.Dict[str, typing.Any]


def _recursive_flatten(to_flatten: typing.Any, strip: str) -> typing.Any:
    if not isinstance(to_flatten, (list, dict)):
        return to_flatten
    if isinstance(to_flatten, list):
        assert isinstance(to_flatten, list)
        return [_recursive_flatten(i, strip) for i in to_flatten]
    assert isinstance(to_flatten, dict)
    keys: typing.List[str] = list(to_flatten.keys())
    copy: str_any_dict = OrderedDict()
    for k in keys:
        item: typing.Any = to_flatten[k]
        if strip in k and strip != k:
            copy[k.split(strip)[1]] = _recursive_flatten(item, strip)
        else:
            copy[k] = _recursive_flatten(item, strip)
    return copy


def flatten_keys(to_flatten: str_any_dict, strip: str) -> str_any_dict:
    keys: typing.List[str] = list(to_flatten.keys())
    copy: str_any_dict = OrderedDict()
    for k in keys:
        item = to_flatten[k]
        if strip in k and strip != k:
            new_key: str = k.split(strip)[1]
            copy[new_key] = _recursive_flatten(item, strip)
        else:
            copy[k] = _recursive_flatten(item, strip)
    return copy


def get_dict_val_case_insensitive(source: typing.Dict[typing.AnyStr, typing.AnyStr],
                                  key: typing.AnyStr) -> typing.Optional[typing.AnyStr]:
    match: typing.List[typing.AnyStr] = list(filter(lambda x: x.lower() == key.lower(), source.keys()))
    if not len(match):
        return None
    if len(match) > 1:
        raise KeyError("overlapping keys")
    if len(match) == 1:
        matched_key: typing.AnyStr = match[0]
        return source[matched_key]
    raise KeyError("overlapping keys")



import asyncio
import time
import typing
import logging
from aioupnp.protocols.scpd import scpd_post
from aioupnp.device import Service
from aioupnp.fault import UPnPError

log = logging.getLogger(__name__)


def soap_optional_str(x: typing.Optional[typing.Union[str, int]]) -> typing.Optional[str]:
    return str(x) if x is not None and str(x).lower() not in ['none', 'nil'] else None


def soap_bool(x: typing.Optional[typing.Union[str, int]]) -> bool:
    return False if not x or str(x).lower() in ['false', 'False'] else True


class GetSpecificPortMappingEntryResponse(typing.NamedTuple):
    internal_port: int
    lan_address: str
    enabled: bool
    description: str
    lease_time: int


class GetGenericPortMappingEntryResponse(typing.NamedTuple):
    gateway_address: str
    external_port: int
    protocol: str
    internal_port: int
    lan_address: str
    enabled: bool
    description: str
    lease_time: int


class SCPDRequestDebuggingInfo(typing.NamedTuple):
    method: str
    kwargs: typing.Dict[str, typing.Union[str, int, bool]]
    response_xml: bytes
    result: typing.Optional[typing.Union[str, int, bool, GetSpecificPortMappingEntryResponse,
                                         GetGenericPortMappingEntryResponse]]
    err: typing.Optional[Exception]
    ts: float


def recast_return(return_annotation, result: typing.Union[str, int, bool, typing.Dict[str, typing.Union[int, str]]],
                  result_keys: typing.List[str]) -> typing.Optional[
                typing.Union[str, int, bool, GetSpecificPortMappingEntryResponse, GetGenericPortMappingEntryResponse]]:
    if len(result_keys) == 1:
        if isinstance(result, (str, int, bool)):
            single_result = result
        else:
            if result_keys[0] in result:
                single_result = result[result_keys[0]]
            else:  # check for the field having incorrect capitalization
                flattened = {k.lower(): v for k, v in result.items()}
                if result_keys[0].lower() in flattened:
                    single_result = flattened[result_keys[0].lower()]
                else:
                    raise UPnPError(f"expected response key {result_keys[0]}, got {list(result.keys())}")
        if return_annotation is bool:
            return soap_bool(single_result)
        if return_annotation is str:
            return soap_optional_str(single_result)
        return None if single_result is None else int(single_result)
    elif return_annotation in [GetGenericPortMappingEntryResponse, GetSpecificPortMappingEntryResponse]:
        assert isinstance(result, dict)
        arg_types: typing.Dict[str, typing.Type[typing.Any]] = return_annotation._field_types
        assert len(arg_types) == len(result_keys)
        recast_results: typing.Dict[str, typing.Optional[typing.Union[str, int, bool]]] = {}
        for i, (field_name, result_key) in enumerate(zip(arg_types, result_keys)):
            result_field_name = result_keys[i]
            field_type = arg_types[field_name]
            if field_type is bool:
                recast_results[field_name] = soap_bool(result.get(result_field_name, None))
            elif field_type is str:
                recast_results[field_name] = soap_optional_str(result.get(result_field_name, None))
            elif field_type is int:
                recast_results[field_name] = int(result[result_field_name]) if result_field_name in result else None
        return return_annotation(**recast_results)
    return None


class SOAPCommands:
    """
    Type annotated wrappers for common UPnP SOAP functions

    A SOAPCommands object has its command attributes overridden during device discovery with SOAPCommand objects
    for the commands implemented by the gateway.

    SOAPCommand will use the typing annotations provided here to properly cast the types of arguments and results
    to their expected types.
    """

    SOAP_COMMANDS: typing.List[str] = [
        'AddPortMapping',
        'GetGenericPortMappingEntry',
        'GetSpecificPortMappingEntry',
        'DeletePortMapping',
        'GetExternalIPAddress',
        # 'SetConnectionType',
        # 'GetNATRSIPStatus',
        # 'GetConnectionTypeInfo',
        # 'GetStatusInfo',
        # 'ForceTermination',
        # 'RequestConnection',
        # 'GetCommonLinkProperties',
        # 'GetTotalBytesSent',
        # 'GetTotalBytesReceived',
        # 'GetTotalPacketsSent',
        # 'GetTotalPacketsReceived',
        # 'X_GetICSStatistics',
        # 'GetDefaultConnectionService',
        # 'SetDefaultConnectionService',
        # 'SetEnabledForInternet',
        # 'GetEnabledForInternet',
        # 'GetMaximumActiveConnections',
        # 'GetActiveConnections'
    ]

    def __init__(self, loop: asyncio.AbstractEventLoop, base_address: bytes, port: int) -> None:
        self._loop = loop
        self._registered: typing.Dict[Service,
                                      typing.Dict[str, typing.Tuple[typing.List[str], typing.List[str]]]] = {}
        self._wrappers_no_args: typing.Dict[str, typing.Callable[[], typing.Awaitable[typing.Any]]] = {}
        self._wrappers_kwargs: typing.Dict[str, typing.Callable[..., typing.Awaitable[typing.Any]]] = {}

        self._base_address = base_address
        self._port = port
        self._request_debug_infos: typing.List[SCPDRequestDebuggingInfo] = []

    def is_registered(self, name: str) -> bool:
        if name not in self.SOAP_COMMANDS:
            raise ValueError("unknown command")
        for service in self._registered.values():
            if name in service:
                return True
        return False

    def get_service(self, name: str) -> Service:
        if name not in self.SOAP_COMMANDS:
            raise ValueError("unknown command")
        for service, commands in self._registered.items():
            if name in commands:
                return service
        raise ValueError(name)

    def _register_soap_wrapper(self, name: str) -> None:
        annotations: typing.Dict[str, typing.Any] = typing.get_type_hints(getattr(self, name))
        service = self.get_service(name)
        input_names: typing.List[str] = self._registered[service][name][0]
        output_names: typing.List[str] = self._registered[service][name][1]

        async def wrapper(**kwargs: typing.Any) -> typing.Optional[
              typing.Union[str, int, bool, GetSpecificPortMappingEntryResponse, GetGenericPortMappingEntryResponse]]:

            assert service.controlURL is not None
            assert service.serviceType is not None
            response, xml_bytes, err = await scpd_post(
                service.controlURL, self._base_address.decode(), self._port, name, input_names,
                service.serviceType.encode(), self._loop, **kwargs
            )
            if err is not None:
                assert isinstance(xml_bytes, bytes)
                self._request_debug_infos.append(SCPDRequestDebuggingInfo(name, kwargs, xml_bytes, None, err, time.time()))
                raise err
            assert 'return' in annotations
            try:
                result = recast_return(annotations['return'], response, output_names)
                self._request_debug_infos.append(SCPDRequestDebuggingInfo(name, kwargs, xml_bytes, result, None, time.time()))
            except Exception as err:
                if isinstance(err, asyncio.CancelledError):
                    raise
                self._request_debug_infos.append(SCPDRequestDebuggingInfo(name, kwargs, xml_bytes, None, err, time.time()))
                raise UPnPError(f"Raised {str(type(err).__name__)}({str(err)}) parsing response for {name}")
            return result

        if not len(list(k for k in annotations if k != 'return')):
            self._wrappers_no_args[name] = wrapper
        else:
            self._wrappers_kwargs[name] = wrapper
        return None

    def register(self, name: str, service: Service, inputs: typing.List[str], outputs: typing.List[str]) -> None:
        if name not in self.SOAP_COMMANDS:
            raise AttributeError(name)
        if self.is_registered(name):
            raise AttributeError(f"{name} is already a registered SOAP command")
        if service not in self._registered:
            self._registered[service] = {}
        self._registered[service][name] = inputs, outputs
        self._register_soap_wrapper(name)

    async def AddPortMapping(self, NewRemoteHost: str, NewExternalPort: int, NewProtocol: str, NewInternalPort: int,
                             NewInternalClient: str, NewEnabled: int, NewPortMappingDescription: str,
                             NewLeaseDuration: str) -> None:
        """Returns None"""
        name = "AddPortMapping"
        if not self.is_registered(name):
            raise NotImplementedError()
        assert name in self._wrappers_kwargs
        await self._wrappers_kwargs[name](
            NewRemoteHost=NewRemoteHost, NewExternalPort=NewExternalPort, NewProtocol=NewProtocol,
            NewInternalPort=NewInternalPort, NewInternalClient=NewInternalClient, NewEnabled=NewEnabled,
            NewPortMappingDescription=NewPortMappingDescription, NewLeaseDuration=NewLeaseDuration
        )
        return None

    async def GetGenericPortMappingEntry(self, NewPortMappingIndex: int) -> GetGenericPortMappingEntryResponse:
        """
        Returns (NewRemoteHost, NewExternalPort, NewProtocol, NewInternalPort, NewInternalClient, NewEnabled,
                 NewPortMappingDescription, NewLeaseDuration)
        """
        name = "GetGenericPortMappingEntry"
        if not self.is_registered(name):
            raise NotImplementedError()
        assert name in self._wrappers_kwargs
        result: GetGenericPortMappingEntryResponse = await self._wrappers_kwargs[name](
            NewPortMappingIndex=NewPortMappingIndex
        )
        return result

    async def GetSpecificPortMappingEntry(self, NewRemoteHost: str, NewExternalPort: int,
                                          NewProtocol: str) -> GetSpecificPortMappingEntryResponse:
        """Returns (NewInternalPort, NewInternalClient, NewEnabled, NewPortMappingDescription, NewLeaseDuration)"""
        name = "GetSpecificPortMappingEntry"
        if not self.is_registered(name):
            raise NotImplementedError()
        assert name in self._wrappers_kwargs
        result: GetSpecificPortMappingEntryResponse = await self._wrappers_kwargs[name](
            NewRemoteHost=NewRemoteHost, NewExternalPort=NewExternalPort, NewProtocol=NewProtocol
        )
        return result

    async def DeletePortMapping(self, NewRemoteHost: str, NewExternalPort: int, NewProtocol: str) -> None:
        """Returns None"""
        name = "DeletePortMapping"
        if not self.is_registered(name):
            raise NotImplementedError()
        assert name in self._wrappers_kwargs
        await self._wrappers_kwargs[name](
            NewRemoteHost=NewRemoteHost, NewExternalPort=NewExternalPort, NewProtocol=NewProtocol
        )
        return None

    async def GetExternalIPAddress(self) -> str:
        """Returns (NewExternalIPAddress)"""
        name = "GetExternalIPAddress"
        if not self.is_registered(name):
            raise NotImplementedError()
        assert name in self._wrappers_no_args
        result: str = await self._wrappers_no_args[name]()
        if not result:
            raise UPnPError("Got null external ip address")
        return result

    # async def GetNATRSIPStatus(self) -> Tuple[bool, bool]:
    #     """Returns (NewRSIPAvailable, NewNATEnabled)"""
    #     name = "GetNATRSIPStatus"
    #     if not self.is_registered(name):
    #         raise NotImplementedError()
    #     assert name in self._wrappers_no_args
    #     result: Tuple[bool, bool] = await self._wrappers_no_args[name]()
    #     return result[0], result[1]
    #
    # async def SetConnectionType(self, NewConnectionType: str) -> None:
    #     """Returns None"""
    #     name = "SetConnectionType"
    #     if not self.is_registered(name):
    #         raise NotImplementedError()
    #     assert name in self._wrappers_kwargs
    #     await self._wrappers_kwargs[name](NewConnectionType=NewConnectionType)
    #     return None
    #
    # async def GetConnectionTypeInfo(self) -> Tuple[str, str]:
    #     """Returns (NewConnectionType, NewPossibleConnectionTypes)"""
    #     name = "GetConnectionTypeInfo"
    #     if not self.is_registered(name):
    #         raise NotImplementedError()
    #     assert name in self._wrappers_no_args
    #     result: Tuple[str, str] = await self._wrappers_no_args[name]()
    #     return result
    #
    # async def GetStatusInfo(self) -> Tuple[str, str, int]:
    #     """Returns (NewConnectionStatus, NewLastConnectionError, NewUptime)"""
    #     name = "GetStatusInfo"
    #     if not self.is_registered(name):
    #         raise NotImplementedError()
    #     assert name in self._wrappers_no_args
    #     result: Tuple[str, str, int] = await self._wrappers_no_args[name]()
    #     return result
    #
    # async def ForceTermination(self) -> None:
    #     """Returns None"""
    #     name = "ForceTermination"
    #     if not self.is_registered(name):
    #         raise NotImplementedError()
    #     assert name in self._wrappers_no_args
    #     await self._wrappers_no_args[name]()
    #     return None
    #
    # async def RequestConnection(self) -> None:
    #     """Returns None"""
    #     name = "RequestConnection"
    #     if not self.is_registered(name):
    #         raise NotImplementedError()
    #     assert name in self._wrappers_no_args
    #     await self._wrappers_no_args[name]()
    #     return None
    #
    # async def GetCommonLinkProperties(self) -> Tuple[str, int, int, str]:
    #     """Returns (NewWANAccessType, NewLayer1UpstreamMaxBitRate, NewLayer1DownstreamMaxBitRate,
    #      NewPhysicalLinkStatus)"""
    #     name = "GetCommonLinkProperties"
    #     if not self.is_registered(name):
    #         raise NotImplementedError()
    #     assert name in self._wrappers_no_args
    #     result: Tuple[str, int, int, str] = await self._wrappers_no_args[name]()
    #     return result
    #
    # async def GetTotalBytesSent(self) -> int:
    #     """Returns (NewTotalBytesSent)"""
    #     name = "GetTotalBytesSent"
    #     if not self.is_registered(name):
    #         raise NotImplementedError()
    #     assert name in self._wrappers_no_args
    #     result: Tuple[int] = await self._wrappers_no_args[name]()
    #     return result[0]
    #
    # async def GetTotalBytesReceived(self) -> int:
    #     """Returns (NewTotalBytesReceived)"""
    #     name = "GetTotalBytesReceived"
    #     if not self.is_registered(name):
    #         raise NotImplementedError()
    #     assert name in self._wrappers_no_args
    #     result: Tuple[int] = await self._wrappers_no_args[name]()
    #     return result[0]
    #
    # async def GetTotalPacketsSent(self) -> int:
    #     """Returns (NewTotalPacketsSent)"""
    #     name = "GetTotalPacketsSent"
    #     if not self.is_registered(name):
    #         raise NotImplementedError()
    #     assert name in self._wrappers_no_args
    #     result: Tuple[int] = await self._wrappers_no_args[name]()
    #     return result[0]
    #
    # async def GetTotalPacketsReceived(self) -> int:
    #     """Returns (NewTotalPacketsReceived)"""
    #     name = "GetTotalPacketsReceived"
    #     if not self.is_registered(name):
    #         raise NotImplementedError()
    #     assert name in self._wrappers_no_args
    #     result: Tuple[int] = await self._wrappers_no_args[name]()
    #     return result[0]
    #
    # async def X_GetICSStatistics(self) -> Tuple[int, int, int, int, str, str]:
    #     """Returns (TotalBytesSent, TotalBytesReceived, TotalPacketsSent, TotalPacketsReceived,
    #     Layer1DownstreamMaxBitRate, Uptime)"""
    #     name = "X_GetICSStatistics"
    #     if not self.is_registered(name):
    #         raise NotImplementedError()
    #     assert name in self._wrappers_no_args
    #     result: Tuple[int, int, int, int, str, str] = await self._wrappers_no_args[name]()
    #     return result
    #
    # async def GetDefaultConnectionService(self) -> str:
    #     """Returns (NewDefaultConnectionService)"""
    #     name = "GetDefaultConnectionService"
    #     if not self.is_registered(name):
    #         raise NotImplementedError()
    #     assert name in self._wrappers_no_args
    #     result: Tuple[str] = await self._wrappers_no_args[name]()
    #     return result[0]
    #
    # async def SetDefaultConnectionService(self, NewDefaultConnectionService: str) -> None:
    #     """Returns (None)"""
    #     name = "SetDefaultConnectionService"
    #     if not self.is_registered(name):
    #         raise NotImplementedError()
    #     assert name in self._wrappers_kwargs
    #     await self._wrappers_kwargs[name](NewDefaultConnectionService=NewDefaultConnectionService)
    #     return None
    #
    # async def SetEnabledForInternet(self, NewEnabledForInternet: bool) -> None:
    #     name = "SetEnabledForInternet"
    #     if not self.is_registered(name):
    #         raise NotImplementedError()
    #     assert name in self._wrappers_kwargs
    #     await self._wrappers_kwargs[name](NewEnabledForInternet=NewEnabledForInternet)
    #     return None
    #
    # async def GetEnabledForInternet(self) -> bool:
    #     name = "GetEnabledForInternet"
    #     if not self.is_registered(name):
    #         raise NotImplementedError()
    #     assert name in self._wrappers_no_args
    #     result: Tuple[bool] = await self._wrappers_no_args[name]()
    #     return result[0]
    #
    # async def GetMaximumActiveConnections(self, NewActiveConnectionIndex: int) -> None:
    #     name = "GetMaximumActiveConnections"
    #     if not self.is_registered(name):
    #         raise NotImplementedError()
    #     assert name in self._wrappers_kwargs
    #     await self._wrappers_kwargs[name](NewActiveConnectionIndex=NewActiveConnectionIndex)
    #     return None
    #
    # async def GetActiveConnections(self) -> Tuple[str, str]:
    #     """Returns (NewActiveConnDeviceContainer, NewActiveConnectionServiceID"""
    #     name = "GetActiveConnections"
    #     if not self.is_registered(name):
    #         raise NotImplementedError()
    #     assert name in self._wrappers_no_args
    #     result: Tuple[str, str] = await self._wrappers_no_args[name]()
    #     return result



import typing
from collections import OrderedDict
from defusedxml import ElementTree

str_any_dict = typing.Dict[str, typing.Any]


def parse_xml(xml_str: str) -> ElementTree:
    element: ElementTree = ElementTree.fromstring(xml_str)
    return element


def _element_text(element_tree: ElementTree) -> typing.Optional[str]:
    # if element_tree.attrib:
    #     element: typing.Dict[str, str] = OrderedDict()
    #     for k, v in element_tree.attrib.items():
    #         element['@' + k] = v
    #     if not element_tree.text:
    #         return element
    #     element['#text'] = element_tree.text.strip()
    #     return element
    if element_tree.text:
        return element_tree.text.strip()
    return None


def _get_element_children(element_tree: ElementTree) -> typing.Dict[str, typing.Union[
                                                     str, typing.List[typing.Union[typing.Dict[str, str], str]]]]:
    element_children = _get_child_dicts(element_tree)
    element: typing.Dict[str, typing.Any] = OrderedDict()
    keys: typing.List[str] = list(element_children.keys())
    for k in keys:
        v: typing.Union[str, typing.List[typing.Any], typing.Dict[str, typing.Any]] = element_children[k]
        if len(v) == 1 and isinstance(v, list):
            l: typing.List[typing.Union[typing.Dict[str, str], str]] = v
            element[k] = l[0]
        else:
            element[k] = v
    return element


def _get_child_dicts(element: ElementTree) -> typing.Dict[str, typing.List[typing.Union[typing.Dict[str, str], str]]]:
    children_dicts: typing.Dict[str, typing.List[typing.Union[typing.Dict[str, str], str]]] = OrderedDict()
    children: typing.List[ElementTree] = list(element)
    for child in children:
        child_dict = _recursive_element_to_dict(child)
        child_keys: typing.List[str] = list(child_dict.keys())
        for k in child_keys:
            assert k in child_dict
            v: typing.Union[typing.Dict[str, str], str] = child_dict[k]
            if k not in children_dicts.keys():
                new_item = [v]
                children_dicts[k] = new_item
            else:
                sublist = children_dicts[k]
                assert isinstance(sublist, list)
                sublist.append(v)
    return children_dicts


def _recursive_element_to_dict(element_tree: ElementTree) -> typing.Dict[str, typing.Any]:
    if len(element_tree):
        element_result: typing.Dict[str, typing.Dict[str, typing.Union[
            str, typing.List[typing.Union[str, typing.Dict[str, typing.Any]]]]]] = OrderedDict()
        children_element = _get_element_children(element_tree)
        if element_tree.tag is not None:
            element_result[element_tree.tag] = children_element
        return element_result
    else:
        element_text = _element_text(element_tree)
        if element_text is not None:
            base_element_result: typing.Dict[str, typing.Any] = OrderedDict()
            if element_tree.tag is not None:
                base_element_result[element_tree.tag] = element_text
            return base_element_result
    null_result: typing.Dict[str, str] = OrderedDict()
    return null_result


def xml_to_dict(xml_str: str) -> typing.Dict[str, typing.Any]:
    return _recursive_element_to_dict(parse_xml(xml_str))



import struct
import socket
import typing
from asyncio.protocols import DatagramProtocol
from asyncio.transports import DatagramTransport
from unittest import mock


def _get_sock(transport: typing.Optional[DatagramTransport]) -> typing.Optional[socket.socket]:
    if transport is None or not hasattr(transport, "_extra"):
        return None
    sock: typing.Optional[socket.socket] = transport.get_extra_info('socket', None)
    assert sock is None or isinstance(sock, (socket.SocketType, mock.MagicMock))
    return sock


class MulticastProtocol(DatagramProtocol):
    def __init__(self, multicast_address: str, bind_address: str) -> None:
        self.multicast_address = multicast_address
        self.bind_address = bind_address
        self.transport: typing.Optional[DatagramTransport] = None

    @property
    def sock(self) -> typing.Optional[socket.socket]:
        return _get_sock(self.transport)

    def get_ttl(self) -> int:
        sock = self.sock
        if sock:
            return sock.getsockopt(socket.IPPROTO_IP, socket.IP_MULTICAST_TTL)
        return 0

    def set_ttl(self, ttl: int = 1) -> None:
        sock = self.sock
        if sock:
            sock.setsockopt(
                socket.IPPROTO_IP, socket.IP_MULTICAST_TTL, struct.pack('b', ttl)
            )
        return None

    def join_group(self, multicast_address: str, bind_address: str) -> None:
        sock = self.sock
        if sock:
            sock.setsockopt(
                socket.IPPROTO_IP, socket.IP_ADD_MEMBERSHIP,
                socket.inet_aton(multicast_address) + socket.inet_aton(bind_address)
            )
        return None

    def leave_group(self, multicast_address: str, bind_address: str) -> None:
        sock = self.sock
        if sock:
            sock.setsockopt(
                socket.IPPROTO_IP, socket.IP_DROP_MEMBERSHIP,
                socket.inet_aton(multicast_address) + socket.inet_aton(bind_address)
            )
        return None

    def connection_made(self, transport: DatagramTransport) -> None:  # type: ignore
        self.transport = transport
        return None

    @classmethod
    def create_multicast_socket(cls, bind_address: str) -> socket.socket:
        sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM, socket.IPPROTO_UDP)
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        sock.bind((bind_address, 0))
        sock.setblocking(False)
        return sock



import logging
import typing
import re
from collections import OrderedDict
import asyncio
from asyncio.protocols import Protocol
from aioupnp.fault import UPnPError
from aioupnp.util import get_dict_val_case_insensitive
from aioupnp.serialization.scpd import deserialize_scpd_get_response
from aioupnp.serialization.scpd import serialize_scpd_get
from aioupnp.serialization.soap import serialize_soap_post, deserialize_soap_post_response


log = logging.getLogger(__name__)


HTTP_CODE_REGEX = re.compile(b"^HTTP[\/]{0,1}1\.[1|0] (\d\d\d)(.*)$")


def parse_http_response_code(http_response: bytes) -> typing.Tuple[bytes, bytes]:
    parsed: typing.List[typing.Tuple[bytes, bytes]] = HTTP_CODE_REGEX.findall(http_response)
    return parsed[0]


def parse_headers(response: bytes) -> typing.Tuple[typing.Dict[bytes, bytes], int, bytes]:
    lines = response.split(b'\r\n')
    headers: typing.Dict[bytes, bytes] = OrderedDict([
        (l.split(b':')[0], b':'.join(l.split(b':')[1:]).lstrip(b' ').rstrip(b' '))
        for l in response.split(b'\r\n')
    ])
    if len(lines) != len(headers):
        raise ValueError("duplicate headers")
    header_keys: typing.List[bytes] = list(headers.keys())
    http_response = header_keys[0]
    response_code, message = parse_http_response_code(http_response)
    del headers[http_response]
    return headers, int(response_code), message


class SCPDHTTPClientProtocol(Protocol):
    """
    This class will make HTTP GET and POST requests

    It differs from spec HTTP in that the version string can be invalid, all we care about is the xml body
    and devices respond with an invalid HTTP version line
    """

    def __init__(self, message: bytes, finished: 'asyncio.Future[typing.Tuple[bytes, bytes, int, bytes]]',
                 soap_method: typing.Optional[str] = None, soap_service_id: typing.Optional[str] = None) -> None:
        self.message = message
        self.response_buff = b""
        self.finished = finished
        self.soap_method = soap_method
        self.soap_service_id = soap_service_id
        self._response_code = 0
        self._response_msg = b""
        self._content_length = 0
        self._got_headers = False
        self._headers: typing.Dict[bytes, bytes] = {}
        self._body = b""
        self.transport: typing.Optional[asyncio.WriteTransport] = None

    def connection_made(self, transport: asyncio.BaseTransport) -> None:
        assert isinstance(transport, asyncio.WriteTransport)
        self.transport = transport
        self.transport.write(self.message)
        return None

    def data_received(self, data: bytes) -> None:
        self.response_buff += data
        for i, line in enumerate(self.response_buff.split(b'\r\n')):
            if not line:  # we hit the blank line between the headers and the body
                if i == (len(self.response_buff.split(b'\r\n')) - 1):
                    return None  # the body is still yet to be written
                if not self._got_headers:
                    self._headers, self._response_code, self._response_msg = parse_headers(
                        b'\r\n'.join(self.response_buff.split(b'\r\n')[:i])
                    )
                    content_length = get_dict_val_case_insensitive(
                        self._headers, b'Content-Length'
                    )
                    if content_length is None:
                        return None
                    self._content_length = int(content_length or 0)
                    self._got_headers = True
                body = b'\r\n'.join(self.response_buff.split(b'\r\n')[i+1:])
                if self._content_length == len(body):
                    self.finished.set_result((self.response_buff, body, self._response_code, self._response_msg))
                elif self._content_length > len(body):
                    pass
                else:
                    self.finished.set_exception(
                        UPnPError(
                            "too many bytes written to response (%i vs %i expected)" % (
                                len(body), self._content_length
                            )
                        )
                    )
                return None
        return None


async def scpd_get(control_url: str, address: str, port: int,
                   loop: typing.Optional[asyncio.AbstractEventLoop] = None) -> typing.Tuple[
                                                     typing.Dict[str, typing.Any], bytes, typing.Optional[Exception]]:
    loop = loop or asyncio.get_event_loop()
    packet = serialize_scpd_get(control_url, address)
    finished: 'asyncio.Future[typing.Tuple[bytes, bytes, int, bytes]]' = loop.create_future()
    proto_factory: typing.Callable[[], SCPDHTTPClientProtocol] = lambda: SCPDHTTPClientProtocol(packet, finished)
    connect_tup: typing.Tuple[asyncio.BaseTransport, asyncio.BaseProtocol] = await loop.create_connection(
        proto_factory, address, port
    )
    protocol = connect_tup[1]
    transport = connect_tup[0]
    assert isinstance(protocol, SCPDHTTPClientProtocol)

    error = None
    wait_task: typing.Awaitable[typing.Tuple[bytes, bytes, int, bytes]] = asyncio.wait_for(protocol.finished, 1.0, loop=loop)
    body = b''
    raw_response = b''
    try:
        raw_response, body, response_code, response_msg = await wait_task
    except asyncio.TimeoutError:
        error = UPnPError("get request timed out")
    except UPnPError as err:
        error = err
        raw_response = protocol.response_buff
    finally:
        transport.close()
    if not error:
        try:
            return deserialize_scpd_get_response(body), raw_response, None
        except Exception as err:
            error = UPnPError(err)

    return {}, raw_response, error


async def scpd_post(control_url: str, address: str, port: int, method: str, param_names: list, service_id: bytes,
                    loop: typing.Optional[asyncio.AbstractEventLoop] = None,
                    **kwargs: typing.Dict[str, typing.Any]
                    ) -> typing.Tuple[typing.Dict, bytes, typing.Optional[Exception]]:
    loop = loop or asyncio.get_event_loop()
    finished: 'asyncio.Future[typing.Tuple[bytes, bytes, int, bytes]]' = loop.create_future()
    packet = serialize_soap_post(method, param_names, service_id, address.encode(), control_url.encode(), **kwargs)
    proto_factory: typing.Callable[[], SCPDHTTPClientProtocol] = lambda:\
        SCPDHTTPClientProtocol(packet, finished, soap_method=method, soap_service_id=service_id.decode())
    connect_tup: typing.Tuple[asyncio.BaseTransport, asyncio.BaseProtocol] = await loop.create_connection(
        proto_factory, address, port
    )
    protocol = connect_tup[1]
    transport = connect_tup[0]
    assert isinstance(protocol, SCPDHTTPClientProtocol)

    try:
        wait_task: typing.Awaitable[typing.Tuple[bytes, bytes, int, bytes]] = asyncio.wait_for(finished, 1.0, loop=loop)
        raw_response, body, response_code, response_msg = await wait_task
    except asyncio.TimeoutError:
        return {}, b'', UPnPError("Timeout")
    except UPnPError as err:
        return {}, protocol.response_buff, err
    finally:
        transport.close()
    try:
        return (
            deserialize_soap_post_response(body, method, service_id.decode()), raw_response, None
        )
    except Exception as err:
        return {}, raw_response, UPnPError(err)



import unittest
from unittest import mock
import socket
import struct
from asyncio import DatagramTransport
from aioupnp.protocols.multicast import MulticastProtocol


class TestMulticast(unittest.TestCase):
    def test_multicast(self):
        _ttl = None
        mock_socket = mock.MagicMock(spec=socket.socket)
        def getsockopt(*_):
            return _ttl

        def setsockopt(a, b, ttl: bytes):
            nonlocal _ttl
            _ttl, = struct.unpack('b', ttl)

        mock_socket.getsockopt = getsockopt
        mock_socket.setsockopt = setsockopt

        protocol = MulticastProtocol('1.2.3.4', '1.2.3.4')
        transport = DatagramTransport()
        transport._extra = {'socket': mock_socket}
        self.assertIsNone(protocol.set_ttl(1))
        self.assertEqual(0, protocol.get_ttl())
        protocol.connection_made(transport)
        self.assertIsNone(protocol.set_ttl(1))
        self.assertEqual(1, protocol.get_ttl())



from collections import OrderedDict
import typing
import logging

log = logging.getLogger(__name__)


class CaseInsensitive:
    def __init__(self, **kwargs: typing.Dict[str, typing.Union[str, typing.Dict[str, typing.Any],
                                                               typing.List[typing.Any]]]) -> None:
        keys: typing.List[str] = list(kwargs.keys())
        for k in keys:
            if not k.startswith("_"):
                assert k in kwargs
                setattr(self, k, kwargs[k])

    def __getattr__(self, item: str) -> typing.Union[str, typing.Dict[str, typing.Any], typing.List]:
        keys: typing.List[str] = list(self.__class__.__dict__.keys())
        for k in keys:
            if k.lower() == item.lower():
                value: typing.Optional[typing.Union[str, typing.Dict[str, typing.Any],
                                                    typing.List]] = self.__dict__.get(k)
                assert value is not None and isinstance(value, (str, dict, list))
                return value
        raise AttributeError(item)

    def __setattr__(self, item: str,
                    value: typing.Union[str, typing.Dict[str, typing.Any], typing.List]) -> None:
        assert isinstance(value, (str, dict)), ValueError(f"got type {str(type(value))}, expected str")
        keys: typing.List[str] = list(self.__class__.__dict__.keys())
        for k in keys:
            if k.lower() == item.lower():
                self.__dict__[k] = value
                return
        if not item.startswith("_"):
            self.__dict__[item] = value
            return
        raise AttributeError(item)

    def as_dict(self) -> typing.Dict[str, typing.Union[str, typing.Dict[str, typing.Any], typing.List]]:
        result: typing.Dict[str, typing.Union[str, typing.Dict[str, typing.Any], typing.List]] = OrderedDict()
        keys: typing.List[str] = list(self.__dict__.keys())
        for k in keys:
            if not k.startswith("_"):
                result[k] = self.__getattr__(k)
        return result


class Service(CaseInsensitive):
    serviceType: typing.Optional[str] = None
    serviceId: typing.Optional[str] = None
    controlURL: typing.Optional[str] = None
    eventSubURL: typing.Optional[str] = None
    SCPDURL: typing.Optional[str] = None


class Device(CaseInsensitive):
    serviceList: typing.Optional[typing.Dict[str, typing.Union[typing.Dict[str, typing.Any], typing.List]]] = None
    deviceList: typing.Optional[typing.Dict[str, typing.Union[typing.Dict[str, typing.Any], typing.List]]] = None
    deviceType: typing.Optional[str] = None
    friendlyName: typing.Optional[str] = None
    manufacturer: typing.Optional[str] = None
    manufacturerURL: typing.Optional[str] = None
    modelDescription: typing.Optional[str] = None
    modelName: typing.Optional[str] = None
    modelNumber: typing.Optional[str] = None
    modelURL: typing.Optional[str] = None
    serialNumber: typing.Optional[str] = None
    udn: typing.Optional[str] = None
    upc: typing.Optional[str] = None
    presentationURL: typing.Optional[str] = None
    iconList: typing.Optional[str] = None

    def __init__(self, devices: typing.List['Device'], services: typing.List[Service],
                 **kwargs: typing.Dict[str, typing.Union[str, typing.Dict[str, typing.Any], typing.List]]) -> None:
        super(Device, self).__init__(**kwargs)
        if self.serviceList and "service" in self.serviceList:
            if isinstance(self.serviceList['service'], dict):
                assert isinstance(self.serviceList['service'], dict)
                svc_list: typing.Dict[str, typing.Any] = self.serviceList['service']
                services.append(Service(**svc_list))
            elif isinstance(self.serviceList['service'], list):
                services.extend(Service(**svc) for svc in self.serviceList["service"])

        if self.deviceList:
            for kw in self.deviceList.values():
                if isinstance(kw, dict):
                    devices.append(Device(devices, services, **kw))
                elif isinstance(kw, list):
                    for _inner_kw in kw:
                        devices.append(Device(devices, services, **_inner_kw))
                else:
                    log.warning("failed to parse device:\n%s", kw)



import sys
import asyncio
import logging
import textwrap
import typing
from collections import OrderedDict
from aioupnp.upnp import run_cli, UPnP, cli_commands

log = logging.getLogger("aioupnp")
handler = logging.StreamHandler()
handler.setFormatter(logging.Formatter('%(asctime)-15s-%(filename)s:%(lineno)s->%(message)s'))
log.addHandler(handler)
log.setLevel(logging.WARNING)

base_usage = "\n".join(textwrap.wrap(
    "aioupnp [-h] [--debug_logging] [--interface=<interface>] [--gateway_address=<gateway_address>]"
    " [--lan_address=<lan_address>] [--timeout=<timeout>] [(--<header_key>=<value>)...]",
    100, subsequent_indent='  ', break_long_words=False)) + "\n"


def get_help(command: str) -> str:
    annotations, doc = UPnP.get_annotations(command)
    doc = doc or ""

    arg_strs = []
    for k, v in annotations.items():
        if k not in ['return', 'igd_args', 'loop']:
            t = str(v) if not hasattr(v, "__name__") else v.__name__
            if t == 'bool':
                arg_strs.append(f"[--{k}]")
            else:
                arg_strs.append(f"[--{k}=<{t}>]")
        elif k == 'igd_args':
            arg_strs.append(f"[--<header key>=<header value>, ...]")

    params = " ".join(arg_strs)
    usage = "\n".join(textwrap.wrap(
        f"aioupnp [-h] [--debug_logging] {command} {params}",
        100, subsequent_indent='  ', break_long_words=False)) + "\n"

    return usage + textwrap.dedent(doc)


def main(argv: typing.Optional[typing.List[typing.Optional[str]]] = None,
         loop: typing.Optional[asyncio.AbstractEventLoop] = None) -> int:
    argv = argv or list(sys.argv)
    help_str = "\n".join(textwrap.wrap(
        " | ".join(cli_commands), 100, initial_indent='  ', subsequent_indent='  ', break_long_words=False
    ))

    usage = \
        "%s\n" \
        "If m-search headers are provided as keyword arguments all of the headers to be used must be provided,\n" \
        "in the order they are to be used. For example:\n" \
        "  aioupnp --HOST=239.255.255.250:1900 --MAN=\"ssdp:discover\" --MX=1 --ST=upnp:rootdevice m_search\n\n" \
        "Commands:\n" \
        "%s\n\n" \
        "For help with a specific command:" \
        "  aioupnp help <command>" % (base_usage, help_str)

    args: typing.List[str] = [str(arg) for arg in argv[1:]]
    if not args:
        print(usage)
        return 0
    if args[0] in ['help', '-h', '--help']:
        if len(args) > 1:
            if args[1].replace("-", "_") in cli_commands:
                print(get_help(args[1].replace("-", "_")))
                return 0
        print(usage)
        return 0

    defaults: typing.Dict[str, typing.Union[bool, str, int]] = {
        'debug_logging': False,
        'interface': 'default',
        'gateway_address': '',
        'lan_address': '',
        'timeout': 3,
    }

    options: typing.Dict[str, typing.Union[bool, str, int]] = OrderedDict()
    command = None
    for arg in args:
        if arg.startswith("--"):
            if "=" in arg:
                k, v = arg.split("=")
                options[k.lstrip('--')] = v
            else:
                options[arg.lstrip('--')] = True
        else:
            command = arg
            break
    if not command:
        print("no command given")
        print(usage)
        return 0
    kwargs = {}
    for arg in args[len(options)+1:]:
        if arg.startswith("--"):
            k, v = arg.split("=")
            k = k.lstrip('--')
            kwargs[k] = v
        else:
            break
    for k in defaults:
        if k not in options:
            options[k] = defaults[k]

    if options.pop('debug_logging'):
        log.setLevel(logging.DEBUG)

    lan_address: str = str(options.pop('lan_address'))
    gateway_address: str = str(options.pop('gateway_address'))
    timeout: int = int(options.pop('timeout'))
    interface: str = str(options.pop('interface'))

    run_cli(
        command.replace('-', '_'), options, lan_address, gateway_address, timeout, interface, kwargs, loop
    )
    return 0


if __name__ == "__main__":
    sys.exit(main())



import re
import logging
import typing
import asyncio
from typing import Dict, List, Optional
from aioupnp.util import get_dict_val_case_insensitive
from aioupnp.constants import SPEC_VERSION, SERVICE
from aioupnp.commands import SOAPCommands
from aioupnp.device import Device, Service
from aioupnp.protocols.ssdp import m_search, multi_m_search
from aioupnp.protocols.scpd import scpd_get
from aioupnp.serialization.ssdp import SSDPDatagram
from aioupnp.util import flatten_keys
from aioupnp.fault import UPnPError

log = logging.getLogger(__name__)

BASE_ADDRESS_REGEX = re.compile("^(http:\/\/\d*\.\d*\.\d*\.\d*:\d*)\/.*$".encode())
BASE_PORT_REGEX = re.compile("^http:\/\/\d*\.\d*\.\d*\.\d*:(\d*)\/.*$".encode())


def get_action_list(element_dict: typing.Dict[str, typing.Union[str, typing.Dict[str, str],
                                                                typing.List[typing.Dict[str, typing.Dict[str, str]]]]]
                    ) -> typing.List[typing.Tuple[str, typing.List[str], typing.List[str]]]:
    service_info = flatten_keys(element_dict, "{%s}" % SERVICE)
    result: typing.List[typing.Tuple[str, typing.List[str], typing.List[str]]] = []
    if "actionList" in service_info:
        action_list = service_info["actionList"]
    else:
        return result
    if not len(action_list):  # it could be an empty string
        return result

    action = action_list["action"]
    if isinstance(action, dict):
        arg_dicts: typing.List[typing.Dict[str, str]] = []
        if not isinstance(action['argumentList']['argument'], list):  # when there is one arg
            arg_dicts.extend([action['argumentList']['argument']])
        else:
            arg_dicts.extend(action['argumentList']['argument'])

        result.append((action_list["action"]['name'], [i['name'] for i in arg_dicts if i['direction'] == 'in'],
                       [i['name'] for i in arg_dicts if i['direction'] == 'out']))
        return result
    assert isinstance(action, list)
    for _action in action:
        if not _action.get('argumentList'):
            result.append((_action['name'], [], []))
        else:
            if not isinstance(_action['argumentList']['argument'], list):  # when there is one arg
                arg_dicts = [_action['argumentList']['argument']]
            else:
                arg_dicts = _action['argumentList']['argument']
            result.append((
                _action['name'],
                [i['name'] for i in arg_dicts if i['direction'] == 'in'],
                [i['name'] for i in arg_dicts if i['direction'] == 'out']
            ))
    return result


def parse_location(location: bytes) -> typing.Tuple[bytes, int]:
    base_address_result: typing.List[bytes] = BASE_ADDRESS_REGEX.findall(location)
    base_address = base_address_result[0]
    port_result: typing.List[bytes] = BASE_PORT_REGEX.findall(location)
    port = int(port_result[0])
    return base_address, port


class Gateway:
    def __init__(self, ok_packet: SSDPDatagram, lan_address: str, gateway_address: str,
                 loop: Optional[asyncio.AbstractEventLoop] = None) -> None:
        self._loop = loop or asyncio.get_event_loop()
        self._ok_packet = ok_packet
        self._lan_address = lan_address
        self.usn: bytes = (ok_packet.usn or '').encode()
        self.ext: bytes = (ok_packet.ext or '').encode()
        self.server: bytes = (ok_packet.server or '').encode()
        self.location: bytes = (ok_packet.location or '').encode()
        self.cache_control: bytes = (ok_packet.cache_control or '').encode()
        self.date: bytes = (ok_packet.date or '').encode()
        self.urn: bytes = (ok_packet.st or '').encode()

        self._xml_response: bytes = b""
        self._service_descriptors: Dict[str, str] = {}

        self.base_address, self.port = parse_location(self.location)
        self.base_ip = self.base_address.lstrip(b"http://").split(b":")[0]
        assert self.base_ip == gateway_address.encode()
        self.path = self.location.split(b"%s:%i/" % (self.base_ip, self.port))[1]

        self.spec_version: Optional[str] = None
        self.url_base: Optional[str] = None

        self._device: Optional[Device] = None
        self._devices: List[Device] = []
        self._services: List[Service] = []

        self._unsupported_actions: Dict[str, typing.List[str]] = {}
        self._registered_commands: Dict[str, str] = {}
        self.commands = SOAPCommands(self._loop, self.base_ip, self.port)

    @property
    def manufacturer_string(self) -> str:
        manufacturer_string = "UNKNOWN GATEWAY"
        if self.devices:
            devices: typing.List[Device] = list(self.devices.values())
            device = devices[0]
            manufacturer_string = f"{device.manufacturer} {device.modelName}"
        return manufacturer_string

    @property
    def services(self) -> Dict[str, Service]:
        services: Dict[str, Service] = {}
        if self._services:
            for service in self._services:
                if service.serviceType is not None:
                    services[service.serviceType] = service
        return services

    @property
    def devices(self) -> Dict[str, Device]:
        devices: Dict[str, Device] = {}
        if self._device:
            for device in self._devices:
                if device.udn is not None:
                    devices[device.udn] = device
        return devices

    # def get_service(self, service_type: str) -> Optional[Service]:
    #     for service in self._services:
    #         if service.serviceType and service.serviceType.lower() == service_type.lower():
    #             return service
    #     return None

    def debug_gateway(self) -> Dict[str, typing.Union[str, bytes, int, Dict, List]]:
        return {
            'manufacturer_string': self.manufacturer_string,
            'gateway_address': self.base_ip.decode(),
            'server': self.server.decode(),
            'urlBase': self.url_base or '',
            'location': self.location.decode(),
            "specVersion": self.spec_version or '',
            'usn': self.usn.decode(),
            'urn': self.urn.decode(),
            'gateway_xml': self._xml_response.decode(),
            'services_xml': self._service_descriptors,
            'services': {service.SCPDURL: service.as_dict() for service in self._services},
            'reply': self._ok_packet.as_dict(),
            'soap_port': self.port,
            'registered_soap_commands': self._registered_commands,
            'unsupported_soap_commands': self._unsupported_actions,
            'soap_requests': list(self.commands._request_debug_infos)
        }

    @classmethod
    async def _try_gateway_from_ssdp(cls, datagram: SSDPDatagram, lan_address: str,
                                     gateway_address: str, 
                                     loop: Optional[asyncio.AbstractEventLoop] = None) -> Optional['Gateway']:
        required_commands: typing.List[str] = [
            'AddPortMapping',
            'DeletePortMapping',
            'GetExternalIPAddress'
        ]
        try:
            gateway = cls(datagram, lan_address, gateway_address, loop=loop)
            log.debug('get gateway descriptor %s', datagram.location)
            await gateway.discover_commands()
            requirements_met = all([gateway.commands.is_registered(required) for required in required_commands])
            if not requirements_met:
                not_met = [
                    required for required in required_commands if not gateway.commands.is_registered(required)
                ]
                assert datagram.location is not None
                log.debug("found gateway %s at %s, but it does not implement required soap commands: %s",
                          gateway.manufacturer_string, gateway.location, not_met)
                return None
            else:
                log.debug('found gateway %s at %s', gateway.manufacturer_string or "device", datagram.location)
                return gateway
        except (asyncio.TimeoutError, UPnPError) as err:
            assert datagram.location is not None
            log.debug("get %s failed (%s), looking for other devices", datagram.location, str(err))
            return None

    @classmethod
    async def _gateway_from_igd_args(cls, lan_address: str, gateway_address: str,
                                     igd_args: typing.Dict[str, typing.Union[int, str]],
                                     timeout: int = 30,
                                     loop: Optional[asyncio.AbstractEventLoop] = None) -> 'Gateway':
        datagram = await m_search(lan_address, gateway_address, igd_args, timeout, loop)
        gateway = await cls._try_gateway_from_ssdp(datagram, lan_address, gateway_address, loop)
        if not gateway:
            raise UPnPError("no gateway found for given args")
        return gateway

    @classmethod
    async def _discover_gateway(cls, lan_address: str, gateway_address: str, timeout: int = 3,
                                loop: Optional[asyncio.AbstractEventLoop] = None) -> 'Gateway':
        ignored: typing.Set[str] = set()
        ssdp_proto = await multi_m_search(
            lan_address, gateway_address, timeout, loop
        )
        try:
            while True:
                datagram = await ssdp_proto.devices.get()
                if datagram.location in ignored:
                    continue
                gateway = await cls._try_gateway_from_ssdp(datagram, lan_address, gateway_address, loop)
                if gateway:
                    return gateway
                elif datagram.location:
                    ignored.add(datagram.location)
        finally:
            ssdp_proto.disconnect()

    @classmethod
    async def discover_gateway(cls, lan_address: str, gateway_address: str, timeout: int = 3,
                               igd_args: Optional[typing.Dict[str, typing.Union[int, str]]] = None,
                               loop: Optional[asyncio.AbstractEventLoop] = None) -> 'Gateway':
        loop = loop or asyncio.get_event_loop()
        if igd_args:
            return await cls._gateway_from_igd_args(lan_address, gateway_address, igd_args, timeout, loop)
        try:
            return await asyncio.wait_for(loop.create_task(
                cls._discover_gateway(lan_address, gateway_address, timeout, loop)
            ), timeout, loop=loop)
        except asyncio.TimeoutError:
            raise UPnPError(f"M-SEARCH for {gateway_address}:1900 timed out")

    async def discover_commands(self) -> None:
        response, xml_bytes, get_err = await scpd_get(
            self.path.decode(), self.base_ip.decode(), self.port, loop=self._loop
        )
        self._xml_response = xml_bytes
        if get_err is not None:
            raise get_err
        spec_version = get_dict_val_case_insensitive(response, SPEC_VERSION)
        if isinstance(spec_version, bytes):
            self.spec_version = spec_version.decode()
        else:
            self.spec_version = spec_version
        url_base = get_dict_val_case_insensitive(response, "urlbase")
        if isinstance(url_base, bytes):
            self.url_base = url_base.decode()
        else:
            self.url_base = url_base
        if not self.url_base:
            self.url_base = self.base_address.decode()
        if response:
            source_keys: typing.List[str] = list(response.keys())
            matches: typing.List[str] = list(filter(lambda x: x.lower() == "device", source_keys))
            match_key = matches[0]
            match: dict = response[match_key]
            # if not len(match):
            #     return None
            # if len(match) > 1:
            #     raise KeyError("overlapping keys")
            # if len(match) == 1:
            #     matched_key: typing.AnyStr = match[0]
            #     return source[matched_key]
            # raise KeyError("overlapping keys")

            self._device = Device(
                self._devices, self._services, **match
            )
        else:
            self._device = Device(self._devices, self._services)
        for service_type in self.services.keys():
            await self.register_commands(self.services[service_type], self._loop)
        return None

    async def register_commands(self, service: Service,
                                loop: Optional[asyncio.AbstractEventLoop] = None) -> None:
        if not service.SCPDURL:
            raise UPnPError("no scpd url")
        if not service.serviceType:
            raise UPnPError("no service type")

        log.debug("get descriptor for %s from %s", service.serviceType, service.SCPDURL)
        service_dict, xml_bytes, get_err = await scpd_get(service.SCPDURL, self.base_ip.decode(), self.port, loop=loop)
        self._service_descriptors[service.SCPDURL] = xml_bytes.decode()

        if get_err is not None:
            log.debug("failed to get descriptor for %s from %s", service.serviceType, service.SCPDURL)
            if xml_bytes:
                log.debug("response: %s", xml_bytes.decode())
            return None
        if not service_dict:
            return None

        action_list = get_action_list(service_dict)
        for name, inputs, outputs in action_list:
            try:
                self.commands.register(name, service, inputs, outputs)
                self._registered_commands[name] = service.serviceType
                log.debug("registered %s::%s", service.serviceType, name)
            except AttributeError:
                self._unsupported_actions.setdefault(service.serviceType, [])
                self._unsupported_actions[service.serviceType].append(name)
                log.debug("available command for %s does not have a wrapper implemented: %s %s %s",
                          service.serviceType, name, inputs, outputs)
            log.debug("registered service %s", service.serviceType)
        return None



import re
import typing
import json
from aioupnp.util import flatten_keys
from aioupnp.fault import UPnPError
from aioupnp.constants import XML_VERSION, ENVELOPE, BODY, FAULT, CONTROL
from aioupnp.serialization.xml import xml_to_dict

CONTENT_NO_XML_VERSION_PATTERN = re.compile(
    "(\<s\:Envelope xmlns\:s=\"http\:\/\/schemas\.xmlsoap\.org\/soap\/envelope\/\"(\s*.)*\>)".encode()
)


def serialize_soap_post(method: str, param_names: typing.List[str], service_id: bytes, gateway_address: bytes,
                        control_url: bytes, **kwargs: typing.Dict[str, str]) -> bytes:
    args = "".join(f"<{n}>{kwargs.get(n)}</{n}>" for n in param_names)
    soap_body = (f'\r\n{XML_VERSION}\r\n<s:Envelope xmlns:s="http://schemas.xmlsoap.org/soap/envelope/" '
                 f's:encodingStyle="http://schemas.xmlsoap.org/soap/encoding/"><s:Body>'
                 f'<u:{method} xmlns:u="{service_id.decode()}">{args}</u:{method}></s:Body></s:Envelope>')
    if "http://" in gateway_address.decode():
        host = gateway_address.decode().split("http://")[1]
    else:
        host = gateway_address.decode()
    return (
        f'POST {control_url.decode()} HTTP/1.1\r\n'  # could be just / even if it shouldn't be
        f'Host: {host}\r\n'
        f'User-Agent: python3/aioupnp, UPnP/1.0, MiniUPnPc/1.9\r\n'
        f'Content-Length: {len(soap_body)}\r\n'
        f'Content-Type: text/xml\r\n'
        f'SOAPAction: \"{service_id.decode()}#{method}\"\r\n'
        f'Connection: Close\r\n'
        f'Cache-Control: no-cache\r\n'
        f'Pragma: no-cache\r\n'
        f'{soap_body}'
        f'\r\n'
    ).encode()


def deserialize_soap_post_response(response: bytes, method: str,
                                   service_id: str) -> typing.Dict[str, typing.Dict[str, str]]:
    parsed: typing.List[typing.List[bytes]] = CONTENT_NO_XML_VERSION_PATTERN.findall(response)
    content = b'' if not parsed else parsed[0][0]
    content_dict = xml_to_dict(content.decode())
    envelope = content_dict[ENVELOPE]
    if not isinstance(envelope[BODY], dict):
        # raise UPnPError('blank response')
        return {}  # TODO: raise
    response_body: typing.Dict[str, typing.Dict[str, typing.Dict[str, str]]] = flatten_keys(
        envelope[BODY], f"{'{' + service_id + '}'}"
    )
    if not response_body:
        # raise UPnPError('blank response')
        return {}  # TODO: raise
    if FAULT in response_body:
        fault: typing.Dict[str, typing.Dict[str, typing.Dict[str, str]]] = flatten_keys(
            response_body[FAULT], "{%s}" % CONTROL
        )
        try:
            raise UPnPError(fault['detail']['UPnPError']['errorDescription'])
        except (KeyError, TypeError, ValueError):
            raise UPnPError(f"Failed to decode error response: {json.dumps(fault)}")
    response_key = None
    for key in response_body:
        if method in key:
            response_key = key
            break
    if not response_key:
        raise UPnPError(f"unknown response fields for {method}: {response_body}")
    return response_body[response_key]



"""
Alleged SSDP discovery documentation

M-SEARCH * HTTP/1.1

Headers
HOST
    Required. Multicast channel and port reserved for SSDP by Internet Assigned Numbers Authority (IANA). Must be
    239.255.255.250:1900. If the port number (“:1900”) is omitted, the receiver should assume the default SSDP port
    number of 1900.
MAN
    Required by HTTP Extension Framework. Unlike the NTS and ST headers, the value of the MAN header is enclosed in
    double quotes; it defines the scope (namespace) of the extension. Must be "ssdp:discover".
MX
    Required. Maximum wait time in seconds. Should be between 1 and 120 inclusive. Device responses should be delayed a
    random duration between 0 and this many seconds to balance load for the control point when it processes responses.
    This value may be increased if a large number of devices are expected to respond. The MX value should not be
    increased to accommodate network characteristics such as latency or propagation delay (for more details, see the
    explanation below). Specified by UPnP vendor. Integer.
ST
    Required. Search Target. Must be one of the following. (cf. NT header in NOTIFY with ssdp:alive above.) Single URI.

    ssdp:all
        Search for all devices and services.

    upnp:rootdevice
        Search for root devices only.

    uuid:device-UUID
        Search for a particular device. Device UUID specified by UPnP vendor.

    urn:schemas-upnp-org:device:deviceType:v
        Search for any device of this type. Device type and version defined by UPnP Forum working committee.

    urn:schemas-upnp-org:service:serviceType:v
        Search for any service of this type. Service type and version defined by UPnP Forum working committee.

    urn:domain-name:device:deviceType:v
        Search for any device of this type. Domain name, device type and version defined by UPnP vendor. Period
        characters in the domain name must be replaced with hyphens in accordance with RFC 2141.

    urn:domain-name:service:serviceType:v
        Search for any service of this type. Domain name, service type and version defined by UPnP vendor. Period
        characters in the domain name must be replaced with hyphens in accordance with RFC 2141.
"""

import typing
from collections import OrderedDict
from aioupnp.constants import SSDP_DISCOVER, SSDP_HOST

SEARCH_TARGETS: typing.List[str] = [
    'upnp:rootdevice',
    'urn:schemas-upnp-org:device:InternetGatewayDevice:1',
    'urn:schemas-wifialliance-org:device:WFADevice:1',
    'urn:schemas-upnp-org:device:WANDevice:1',
    "urn:schemas-upnp-org:service:WANIPConnection:1",
    "urn:schemas-upnp-org:service:WANPPPConnection:1",
    'ssdp:all'
]


def format_packet_args(order: typing.List[str],
                       kwargs: typing.Dict[str, typing.Union[int, str]]) -> typing.Dict[str, typing.Union[int, str]]:
    args = []
    for o in order:
        for k, v in kwargs.items():
            if k.lower() == o.lower():
                args.append((k, v))
                break
    return OrderedDict(args)


def packet_generator() -> typing.Iterator[typing.Dict[str, typing.Union[int, str]]]:
    for st in SEARCH_TARGETS:
        order = ["HOST", "MAN", "MX", "ST"]
        yield format_packet_args(order, {'HOST': SSDP_HOST, 'MAN': '"%s"' % SSDP_DISCOVER, 'MX': 1, 'ST': st})
        yield format_packet_args(order, {'Host': SSDP_HOST, 'Man': '"%s"' % SSDP_DISCOVER, 'MX': 1, 'ST': st})
        yield format_packet_args(order, {'HOST': SSDP_HOST, 'MAN': SSDP_DISCOVER, 'MX': 1, 'ST': st})
        yield format_packet_args(order, {'Host': SSDP_HOST, 'Man': SSDP_DISCOVER, 'MX': 1, 'ST': st})

        order = ["HOST", "MAN", "ST", "MX"]
        yield format_packet_args(order, {'HOST': SSDP_HOST, 'MAN': '"%s"' % SSDP_DISCOVER, 'MX': 1, 'ST': st})
        yield format_packet_args(order, {'HOST': SSDP_HOST, 'MAN': SSDP_DISCOVER, 'MX': 1, 'ST': st})

        order = ["HOST", "ST", "MAN", "MX"]
        yield format_packet_args(order, {'HOST': SSDP_HOST, 'MAN': '"%s"' % SSDP_DISCOVER, 'MX': 1, 'ST': st})
        yield format_packet_args(order, {'HOST': SSDP_HOST, 'MAN': SSDP_DISCOVER, 'MX': 1, 'ST': st})



from collections import OrderedDict
from aioupnp.fault import UPnPError
from aioupnp.protocols.m_search_patterns import packet_generator
from aioupnp.serialization.ssdp import SSDPDatagram
from aioupnp.constants import SSDP_IP_ADDRESS
from aioupnp.protocols.ssdp import m_search, SSDPProtocol
from tests import AsyncioTestCase, mock_tcp_and_udp


class TestSSDP(AsyncioTestCase):
    packet_args = list(packet_generator())
    byte_packets = [SSDPDatagram("M-SEARCH", p).encode().encode() for p in packet_args]

    successful_args = OrderedDict([
        ("HOST", "239.255.255.250:1900"),
        ("MAN", "ssdp:discover"),
        ("MX", 1),
        ("ST", "urn:schemas-upnp-org:device:WANDevice:1")
    ])
    query_packet = SSDPDatagram("M-SEARCH", successful_args)

    reply_args = OrderedDict([
        ("CACHE_CONTROL", "max-age=1800"),
        ("LOCATION", "http://10.0.0.1:49152/InternetGatewayDevice.xml"),
        ("SERVER", "Linux, UPnP/1.0, DIR-890L Ver 1.20"),
        ("ST", "urn:schemas-upnp-org:device:WANDevice:1"),
        ("USN", "uuid:22222222-3333-4444-5555-666666666666::urn:schemas-upnp-org:device:WANDevice:1")
    ])
    reply_packet = SSDPDatagram("OK", reply_args)

    async def test_socket_setup_error(self):
        with mock_tcp_and_udp(self.loop, raise_oserror_on_bind=True):
            with self.assertRaises(UPnPError):
                await m_search("10.0.0.2", "10.0.0.1", self.successful_args, timeout=1, loop=self.loop)

    async def test_transport_not_connected_error(self):
        try:
            await SSDPProtocol('', '').m_search('1.2.3.4', 2,  [self.query_packet.as_dict()])
            self.assertTrue(False)
        except UPnPError as err:
            self.assertEqual(str(err), "SSDP transport not connected")

    async def test_deadbeef_response(self):
        replies = {
            (self.query_packet.encode().encode(), ("10.0.0.1", 1900)): b'\xde\xad\xbe\xef'
        }
        sent = []

        with mock_tcp_and_udp(self.loop, udp_replies=replies, udp_expected_addr="10.0.0.1", sent_udp_packets=sent):
            with self.assertRaises(UPnPError):
                await m_search("10.0.0.2", "10.0.0.1", self.successful_args, timeout=1, loop=self.loop)

    async def test_m_search_reply_unicast(self):
        replies = {
            (self.query_packet.encode().encode(), ("10.0.0.1", 1900)): self.reply_packet.encode().encode()
        }
        sent = []

        with mock_tcp_and_udp(self.loop, udp_replies=replies, udp_expected_addr="10.0.0.1", sent_udp_packets=sent):
            reply = await m_search("10.0.0.2", "10.0.0.1", self.successful_args, timeout=1, loop=self.loop)

        self.assertEqual(reply.encode(), self.reply_packet.encode())
        self.assertIn(self.query_packet.encode().encode(), sent)

        with self.assertRaises(UPnPError):
            with mock_tcp_and_udp(self.loop, udp_expected_addr="10.0.0.10", udp_replies=replies):
                await m_search("10.0.0.2", "10.0.0.1", self.successful_args, timeout=1, loop=self.loop)

    async def test_m_search_reply_multicast(self):
        replies = {
            (self.query_packet.encode().encode(), (SSDP_IP_ADDRESS, 1900)): self.reply_packet.encode().encode()
        }
        sent = []

        with mock_tcp_and_udp(self.loop, udp_replies=replies, udp_expected_addr="10.0.0.1", sent_udp_packets=sent):
            reply = await m_search("10.0.0.2", "10.0.0.1", self.successful_args, timeout=1, loop=self.loop)

        self.assertEqual(reply.encode(), self.reply_packet.encode())
        self.assertIn(self.query_packet.encode().encode(), sent)

        with self.assertRaises(UPnPError):
            with mock_tcp_and_udp(self.loop, udp_replies=replies, udp_expected_addr="10.0.0.10"):
                await m_search("10.0.0.2", "10.0.0.1", self.successful_args, timeout=1, loop=self.loop)

    # async def test_packets_sent_fuzzy_m_search(self):
    #     sent = []
    #
    #     with self.assertRaises(UPnPError):
    #         with mock_tcp_and_udp(self.loop, udp_expected_addr="10.0.0.1", sent_udp_packets=sent):
    #             await fuzzy_m_search("10.0.0.2", "10.0.0.1", 1, self.loop)
    #     for packet in self.byte_packets:
    #         self.assertIn(packet, sent)
    #
    # async def test_packets_fuzzy_m_search(self):
    #     replies = {
    #         (self.query_packet.encode().encode(), (SSDP_IP_ADDRESS, 1900)): self.reply_packet.encode().encode()
    #     }
    #     sent = []
    #
    #     with mock_tcp_and_udp(self.loop, udp_expected_addr="10.0.0.1", udp_replies=replies, sent_udp_packets=sent):
    #         args, reply = await fuzzy_m_search("10.0.0.2", "10.0.0.1", 1, self.loop)
    #
    #     self.assertEqual(reply.encode(), self.reply_packet.encode())
    #     self.assertEqual(args, self.successful_args)
    #
    # async def test_packets_sent_fuzzy_m_search_ignore_invalid_datagram_replies(self):
    #     sent = []
    #
    #     with self.assertRaises(UPnPError):
    #         with mock_tcp_and_udp(self.loop, udp_expected_addr="10.0.0.1", sent_udp_packets=sent,
    #                               add_potato_datagrams=True):
    #             await fuzzy_m_search("10.0.0.2", "10.0.0.1", 1, self.loop)
    #
    #     for packet in self.byte_packets:
    #         self.assertIn(packet, sent)


# import os
# import zlib
# import base64
import logging
import json
import asyncio
from typing import Tuple, Dict, List, Union, Optional, Any
from aioupnp.fault import UPnPError
from aioupnp.gateway import Gateway
from aioupnp.interfaces import get_gateway_and_lan_addresses
from aioupnp.serialization.ssdp import SSDPDatagram
from aioupnp.commands import GetGenericPortMappingEntryResponse, GetSpecificPortMappingEntryResponse


log = logging.getLogger(__name__)


# def _encode(x):
#     if isinstance(x, bytes):
#         return x.decode()
#     elif isinstance(x, Exception):
#         return str(x)
#     return x


class UPnP:
    def __init__(self, lan_address: str, gateway_address: str, gateway: Gateway) -> None:
        self.lan_address = lan_address
        self.gateway_address = gateway_address
        self.gateway = gateway

    @classmethod
    def get_annotations(cls, command: str) -> Tuple[Dict[str, Any], Optional[str]]:
        if command == "m_search":
            return cls.m_search.__annotations__, cls.m_search.__doc__
        if command == "get_external_ip":
            return cls.get_external_ip.__annotations__, cls.get_external_ip.__doc__
        if command == "add_port_mapping":
            return cls.add_port_mapping.__annotations__, cls.add_port_mapping.__doc__
        if command == "get_port_mapping_by_index":
            return cls.get_port_mapping_by_index.__annotations__, cls.get_port_mapping_by_index.__doc__
        if command == "get_redirects":
            return cls.get_redirects.__annotations__, cls.get_redirects.__doc__
        if command == "get_specific_port_mapping":
            return cls.get_specific_port_mapping.__annotations__, cls.get_specific_port_mapping.__doc__
        if command == "delete_port_mapping":
            return cls.delete_port_mapping.__annotations__, cls.delete_port_mapping.__doc__
        if command == "get_next_mapping":
            return cls.get_next_mapping.__annotations__, cls.get_next_mapping.__doc__
        raise AttributeError(command)

    @staticmethod
    def get_lan_and_gateway(lan_address: str = '', gateway_address: str = '',
                            interface_name: str = 'default') -> Tuple[str, str]:
        if not lan_address or not gateway_address:
            gateway_addr, lan_addr = get_gateway_and_lan_addresses(interface_name)
            lan_address = lan_address or lan_addr
            gateway_address = gateway_address or gateway_addr
        return lan_address, gateway_address

    @classmethod
    async def discover(cls, lan_address: str = '', gateway_address: str = '', timeout: int = 3,
                       igd_args: Optional[Dict[str, Union[str, int]]] = None, interface_name: str = 'default',
                       loop: Optional[asyncio.AbstractEventLoop] = None) -> 'UPnP':
        lan_address, gateway_address = cls.get_lan_and_gateway(lan_address, gateway_address, interface_name)
        gateway = await Gateway.discover_gateway(
            lan_address, gateway_address, timeout, igd_args, loop
        )
        return cls(lan_address, gateway_address, gateway)

    @classmethod
    async def m_search(cls, lan_address: str = '', gateway_address: str = '', timeout: int = 1,
                       interface_name: str = 'default',
                       igd_args: Optional[Dict[str, Union[str, int]]] = None,
                       loop: Optional[asyncio.AbstractEventLoop] = None
                       ) -> Dict[str, Union[str, Dict[str, Union[str, int]]]]:
        """
        Perform a M-SEARCH for a upnp gateway.

        :param lan_address: (str) the local interface ipv4 address
        :param gateway_address: (str) the gateway ipv4 address
        :param timeout: (int) m search timeout
        :param interface_name: (str) name of the network interface
        :param igd_args: (dict) case sensitive M-SEARCH headers. if used all headers to be used must be provided.

        :return: {
            'lan_address': (str) lan address,
            'gateway_address': (str) gateway address,
            'm_search_kwargs': (str) equivalent igd_args ,
            'discover_reply': (dict) SSDP response datagram
        }
        """

        if not lan_address or not gateway_address:
            try:
                lan_address, gateway_address = cls.get_lan_and_gateway(lan_address, gateway_address, interface_name)
                assert gateway_address and lan_address
            except Exception as err:
                raise UPnPError("failed to get lan and gateway addresses for interface \"%s\": %s" % (interface_name,
                                                                                                      str(err)))
        gateway = await Gateway.discover_gateway(
            lan_address, gateway_address, timeout, igd_args, loop
        )
        return {
            'lan_address': lan_address,
            'gateway_address': gateway_address,
            # 'm_search_kwargs': SSDPDatagram("M-SEARCH", igd_args).get_cli_igd_kwargs(),
            'discover_reply': gateway._ok_packet.as_dict()
        }

    async def get_external_ip(self) -> str:
        """
        Get the external ip address from the gateway

        :return: (str) external ip
        """
        return await self.gateway.commands.GetExternalIPAddress()

    async def add_port_mapping(self, external_port: int, protocol: str, internal_port: int, lan_address: str,
                               description: str) -> None:
        """
        Add a new port mapping

        :param external_port: (int) external port to map
        :param protocol: (str) UDP | TCP
        :param internal_port: (int) internal port
        :param lan_address: (str) internal lan address
        :param description: (str) mapping description
        :return: None
        """
        await self.gateway.commands.AddPortMapping(
            NewRemoteHost='', NewExternalPort=external_port, NewProtocol=protocol,
            NewInternalPort=internal_port, NewInternalClient=lan_address,
            NewEnabled=1, NewPortMappingDescription=description, NewLeaseDuration='0'
        )
        return None

    async def get_port_mapping_by_index(self, index: int) -> GetGenericPortMappingEntryResponse:
        """
        Get information about a port mapping by index number

        :param index: (int) mapping index number
        :return: NamedTuple[
            gateway_address: str
            external_port: int
            protocol: str
            internal_port: int
            lan_address: str
            enabled: bool
            description: str
            lease_time: int
        ]
        """
        return await self.gateway.commands.GetGenericPortMappingEntry(NewPortMappingIndex=index)

    async def get_redirects(self) -> List[GetGenericPortMappingEntryResponse]:
        """
        Get information about all mapped ports

        :return: List[
            NamedTuple[
                gateway_address: str
                external_port: int
                protocol: str
                internal_port: int
                lan_address: str
                enabled: bool
                description: str
                lease_time: int
            ]
        ]
        """
        redirects: List[GetGenericPortMappingEntryResponse] = []
        cnt = 0
        try:
            redirect = await self.get_port_mapping_by_index(cnt)
        except UPnPError:
            return redirects
        while redirect is not None:
            redirects.append(redirect)
            cnt += 1
            try:
                redirect = await self.get_port_mapping_by_index(cnt)
            except UPnPError:
                break
        return redirects

    async def get_specific_port_mapping(self, external_port: int, protocol: str) -> GetSpecificPortMappingEntryResponse:
        """
        Get information about a port mapping by port number and protocol

        :param external_port: (int) port number
        :param protocol: (str) UDP | TCP
        :return: NamedTuple[
            internal_port: int
            lan_address: str
            enabled: bool
            description: str
            lease_time: int
        ]
        """
        return await self.gateway.commands.GetSpecificPortMappingEntry(
            NewRemoteHost='', NewExternalPort=external_port, NewProtocol=protocol
        )

    async def delete_port_mapping(self, external_port: int, protocol: str) -> None:
        """
        Delete a port mapping

        :param external_port: (int) port number of mapping
        :param protocol: (str) TCP | UDP
        :return: None
        """
        await self.gateway.commands.DeletePortMapping(
            NewRemoteHost="", NewExternalPort=external_port, NewProtocol=protocol
        )

        return None

    async def get_next_mapping(self, port: int, protocol: str, description: str,
                               internal_port: Optional[int] = None) -> int:
        """
        Get a new port mapping. If the requested port is not available, increment until the next free port is mapped

        :param port: (int) external port
        :param protocol: (str) UDP | TCP
        :param description: (str) mapping description
        :param internal_port: (int) internal port

        :return: (int) mapped port
        """

        _internal_port = int(internal_port or port)
        requested_port = int(_internal_port)
        port = int(port)
        redirect_tups = await self.get_redirects()

        redirects: Dict[Tuple[int, str], Tuple[str, int, str]] = {
            (ext_port, proto): (int_host, int_port, desc)
            for (ext_host, ext_port, proto, int_port, int_host, enabled, desc, _) in redirect_tups
        }

        while (port, protocol) in redirects:
            int_host, int_port, desc = redirects[(port, protocol)]
            if int_host == self.lan_address and int_port == requested_port and desc == description:
                return port
            port += 1
        await self.add_port_mapping(port, protocol, _internal_port, self.lan_address, description)
        return port

    # @cli
    # async def debug_gateway(self) -> str:
    #     return json.dumps({
    #         "gateway": self.gateway.debug_gateway(),
    #         "client_address": self.lan_address,
    #     }, default=_encode, indent=2)
    #
    # @property
    # def zipped_debugging_info(self) -> str:
    #     return base64.b64encode(zlib.compress(
    #         json.dumps({
    #             "gateway": self.gateway.debug_gateway(),
    #             "client_address": self.lan_address,
    #         }, default=_encode, indent=2).encode()
    #     )).decode()
    #
    # @cli
    # async def get_natrsip_status(self) -> Tuple[bool, bool]:
    #     """Returns (NewRSIPAvailable, NewNATEnabled)"""
    #     return await self.gateway.commands.GetNATRSIPStatus()
    #
    # @cli
    # async def set_connection_type(self, NewConnectionType: str) -> None:
    #     """Returns None"""
    #     return await self.gateway.commands.SetConnectionType(NewConnectionType)
    #
    # @cli
    # async def get_connection_type_info(self) -> Tuple[str, str]:
    #     """Returns (NewConnectionType, NewPossibleConnectionTypes)"""
    #     return await self.gateway.commands.GetConnectionTypeInfo()
    #
    # @cli
    # async def get_status_info(self) -> Tuple[str, str, int]:
    #     """Returns (NewConnectionStatus, NewLastConnectionError, NewUptime)"""
    #     return await self.gateway.commands.GetStatusInfo()
    #
    # @cli
    # async def force_termination(self) -> None:
    #     """Returns None"""
    #     return await self.gateway.commands.ForceTermination()
    #
    # @cli
    # async def request_connection(self) -> None:
    #     """Returns None"""
    #     return await self.gateway.commands.RequestConnection()
    #
    # @cli
    # async def get_common_link_properties(self):
    #     """Returns (NewWANAccessType, NewLayer1UpstreamMaxBitRate, NewLayer1DownstreamMaxBitRate,
    #      NewPhysicalLinkStatus)"""
    #     return await self.gateway.commands.GetCommonLinkProperties()
    #
    # @cli
    # async def get_total_bytes_sent(self) -> int:
    #     """Returns (NewTotalBytesSent)"""
    #     return await self.gateway.commands.GetTotalBytesSent()
    #
    # @cli
    # async def get_total_bytes_received(self):
    #     """Returns (NewTotalBytesReceived)"""
    #     return await self.gateway.commands.GetTotalBytesReceived()
    #
    # @cli
    # async def get_total_packets_sent(self):
    #     """Returns (NewTotalPacketsSent)"""
    #     return await self.gateway.commands.GetTotalPacketsSent()
    #
    # @cli
    # async def get_total_packets_received(self):
    #     """Returns (NewTotalPacketsReceived)"""
    #     return await self.gateway.commands.GetTotalPacketsReceived()
    #
    # @cli
    # async def x_get_ics_statistics(self) -> Tuple[int, int, int, int, str, str]:
    #     """Returns (TotalBytesSent, TotalBytesReceived, TotalPacketsSent, TotalPacketsReceived,
    #      Layer1DownstreamMaxBitRate, Uptime)"""
    #     return await self.gateway.commands.X_GetICSStatistics()
    #
    # @cli
    # async def get_default_connection_service(self):
    #     """Returns (NewDefaultConnectionService)"""
    #     return await self.gateway.commands.GetDefaultConnectionService()
    #
    # @cli
    # async def set_default_connection_service(self, NewDefaultConnectionService: str) -> None:
    #     """Returns (None)"""
    #     return await self.gateway.commands.SetDefaultConnectionService(NewDefaultConnectionService)
    #
    # @cli
    # async def set_enabled_for_internet(self, NewEnabledForInternet: bool) -> None:
    #     return await self.gateway.commands.SetEnabledForInternet(NewEnabledForInternet)
    #
    # @cli
    # async def get_enabled_for_internet(self) -> bool:
    #     return await self.gateway.commands.GetEnabledForInternet()
    #
    # @cli
    # async def get_maximum_active_connections(self, NewActiveConnectionIndex: int):
    #     return await self.gateway.commands.GetMaximumActiveConnections(NewActiveConnectionIndex)
    #
    # @cli
    # async def get_active_connections(self) -> Tuple[str, str]:
    #     """Returns (NewActiveConnDeviceContainer, NewActiveConnectionServiceID"""
    #     return await self.gateway.commands.GetActiveConnections()


cli_commands = [
    'm_search',
    'get_external_ip',
    'add_port_mapping',
    'get_port_mapping_by_index',
    'get_redirects',
    'get_specific_port_mapping',
    'delete_port_mapping',
    'get_next_mapping'
]


def run_cli(method: str, igd_args: Dict[str, Union[bool, str, int]], lan_address: str = '',
            gateway_address: str = '', timeout: int = 3, interface_name: str = 'default',
            kwargs: Optional[Dict[str, str]] = None,
            loop: Optional[asyncio.AbstractEventLoop] = None) -> None:

    kwargs = kwargs or {}
    igd_args = igd_args
    timeout = int(timeout)
    loop = loop or asyncio.get_event_loop()
    fut: 'asyncio.Future' = loop.create_future()

    async def wrapper():  # wrap the upnp setup and call of the command in a coroutine
        if method == 'm_search':  # if we're only m_searching don't do any device discovery
            fn = lambda *_a, **_kw: UPnP.m_search(
                lan_address, gateway_address, timeout, interface_name, igd_args, loop
            )
        else:  # automatically discover the gateway
            try:
                u = await UPnP.discover(
                    lan_address, gateway_address, timeout, igd_args, interface_name, loop=loop
                )
            except UPnPError as err:
                fut.set_exception(err)
                return
            if method not in cli_commands:
                fut.set_exception(UPnPError("\"%s\" is not a recognized command" % method))
                return
            else:
                fn = getattr(u, method)

        try:  # call the command
            result = await fn(**{k: fn.__annotations__[k](v) for k, v in kwargs.items()})
            fut.set_result(result)
        except UPnPError as err:
            fut.set_exception(err)

        except Exception as err:
            log.exception("uncaught error")
            fut.set_exception(UPnPError("uncaught error: %s" % str(err)))

    if not hasattr(UPnP, method):
        fut.set_exception(UPnPError("\"%s\" is not a recognized command" % method))
    else:
        loop.run_until_complete(wrapper())
    try:
        result = fut.result()
    except UPnPError as err:
        print("aioupnp encountered an error: %s" % str(err))
        return

    if isinstance(result, (list, tuple, dict)):
        print(json.dumps(result, indent=2))
    else:
        print(result)
    return
